"""Definition of the Bond class.

This module defines the Bond object in the Batoms package.

"""

import bpy
from time import time
from batoms.utils.butils import object_mode, get_node_by_name
from batoms.utils import string2Number, number2String
import numpy as np
from batoms.base.object import ObjectGN
from batoms.base.collection import BaseCollection
from .setting import BondSettings
from batoms.bond.search_bond import SearchBond, default_search_bond_datas

import logging

logger = logging.getLogger(__name__)


default_bond_attributes = [
    {"name": "atoms_index0", "data_type": "INT", "domain": "EDGE"},
    {"name": "atoms_index1", "data_type": "INT", "domain": "EDGE"},
    {"name": "atoms_index2", "data_type": "INT", "domain": "EDGE"},
    {"name": "atoms_index3", "data_type": "INT", "domain": "EDGE"},
    {"name": "species_index0", "data_type": "INT", "domain": "EDGE"},
    {"name": "species_index1", "data_type": "INT", "domain": "EDGE"},
    {"name": "bond_offset0", "data_type": "FLOAT_VECTOR", "domain": "EDGE"},
    {"name": "bond_offset1", "data_type": "FLOAT_VECTOR", "domain": "EDGE"},
    {"name": "bond_offset2", "data_type": "FLOAT_VECTOR", "domain": "EDGE"},
    {"name": "bond_offset3", "data_type": "FLOAT_VECTOR", "domain": "EDGE"},
    {"name": "bond_order", "data_type": "INT", "domain": "EDGE"},
    {"name": "bond_style", "data_type": "INT", "domain": "EDGE"},
    {"name": "bond_show", "data_type": "INT", "domain": "EDGE"},
    {"name": "bond_model_style", "data_type": "INT", "domain": "EDGE"},
    {"name": "polyhedra", "data_type": "INT", "domain": "EDGE"},
    {"name": "second_bond", "data_type": "INT", "domain": "EDGE"},
]

default_bond_datas = {
    "atoms_index0": np.ones(0, dtype=int),
    "atoms_index1": np.ones(0, dtype=int),
    "atoms_index2": np.ones(0, dtype=int),
    "atoms_index3": np.ones(0, dtype=int),
    "species_index0": np.ones(0, dtype=int),
    "species_index1": np.ones(0, dtype=int),
    "centers": np.zeros((1, 0, 3)),
    # 'vectors':np.zeros((0, 3)),
    "bond_offset0": np.zeros((0, 3)),
    "bond_offset1": np.zeros((0, 3)),
    "bond_offset2": np.zeros((0, 3)),
    "bond_offset3": np.zeros((0, 3)),
    # 'eulers':np.eye(3),
    # 'lengths':np.zeros((0, 3)),
    # 'widths': np.ones(0, dtype=float),
    "bond_show": np.zeros(0, dtype=int),
    "bond_order": np.zeros(0, dtype=int),
    "bond_style": np.zeros(0, dtype=int),
    "bond_model_style": np.ones(0, dtype=int),
    "polyhedra": np.ones(0, dtype=float),
    "second_bond": np.ones(0, dtype=int),
}


class Bond(BaseCollection, ObjectGN):
    """Bbond Class

    A Bbond object is linked to this main collection in Blender.

    Parameters:


    """

    def __init__(
        self,
        label=None,
        batoms=None,
        bond_datas=None,
        location=np.array([0, 0, 0]),
    ):
        #
        self.batoms = batoms
        self.label = label
        ObjectGN.__init__(self, label)
        BaseCollection.__init__(self, coll_name=label)
        self.settings = BondSettings(self.label, batoms=batoms, bonds=self)
        self.build_geometry_node()
        self._search_bond = SearchBond(self.label, batoms=batoms, load=True)

    @property
    def bond_node(self):
        """Get the top level node of bond node group."""
        from batoms.utils.butils import get_node_with_node_tree_by_name

        default_interface = [
            ["Geometry", "NodeSocketGeometry", "INPUT"],
            ["Geometry", "NodeSocketGeometry", "OUTPUT"],
        ]
        name = f"Bond_{self.label}"
        node = get_node_with_node_tree_by_name(
            self.gn_node_group.nodes, name=name, interface=default_interface
        )
        return node

    def get_bond_pair_node(self, name):
        """Get the node of bond pair node group.
        Create the node if not exist."""
        from batoms.utils.butils import get_node_with_node_tree_by_name

        # create group input and output sockets
        default_interface = [
            ["Geometry", "NodeSocketGeometry", "INPUT"],
            ["Species_index0", "NodeSocketInt", "INPUT"],
            ["Species0", "NodeSocketInt", "INPUT"],
            ["Species_index1", "NodeSocketInt", "INPUT"],
            ["Species1", "NodeSocketInt", "INPUT"],
            ["Model_style0", "NodeSocketInt", "INPUT"],
            ["Model_style1", "NodeSocketInt", "INPUT"],
            ["Order", "NodeSocketInt", "INPUT"],
            ["Style", "NodeSocketInt", "INPUT"],
            ["Length", "NodeSocketFloat", "INPUT"],
            ["Rotation", "NodeSocketVector", "INPUT"],
            ["Geometry", "NodeSocketGeometry", "OUTPUT"],
        ]
        # Create Node Group if not exit
        node = get_node_with_node_tree_by_name(
            self.bond_node.node_tree.nodes, name=name, interface=default_interface
        )
        return node

    def build_geometry_node(self):
        """
        v = (p1 + o1) - (p2 + o2)
        len(v) < max_length
        v align euler
        """
        from batoms.utils.butils import get_node_by_name, get_socket_by_identifier

        tstart = time()
        parent = self.gn_node_group
        node = self.bond_node
        nodes = node.node_tree.nodes
        links = node.node_tree.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # link the input to parent node
        JoinGeometry = get_node_by_name(
            parent.nodes,
            "%s_JoinGeometry" % self.batoms.label,
            "GeometryNodeJoinGeometry",
        )
        parent.links.new(
            parent.nodes["Group Input"].outputs["Geometry"], node.inputs["Geometry"]
        )
        # link the outputs to parent node
        parent.links.new(node.outputs["Geometry"], JoinGeometry.inputs["Geometry"])
        # ------------------------------------------------------------------
        # this join geometry is belong to this node group
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        links.new(JoinGeometry.outputs["Geometry"], GroupOutput.inputs["Geometry"])
        # ------------------------------------------------------------------
        self.build_bond_mesh()
        # get the species by atoms_index
        # order
        SpeciesIndexAttribute = get_node_by_name(
            nodes,
            "%s_NamedAttribute_species_index" % (self.label),
            "GeometryNodeInputNamedAttribute",
        )
        SpeciesIndexAttribute.inputs["Name"].default_value = "species_index"
        SpeciesIndexAttribute.data_type = "INT"
        SpeciesAtIndexs = []
        for i in range(2):
            AtomsIndexAttribute = get_node_by_name(
                nodes,
                "%s_NamedAttribute_atoms_index%s" % (self.label, i),
                "GeometryNodeInputNamedAttribute",
            )
            AtomsIndexAttribute.inputs["Name"].default_value = f"atoms_index{i}"
            AtomsIndexAttribute.data_type = "INT"
            SpeciesAtIndex = get_node_by_name(
                nodes,
                "%s_SpeciesAtIndex%s" % (self.label, i),
                "GeometryNodeSampleIndex",
            )
            SpeciesAtIndex.data_type = "INT"
            SpeciesAtIndexs.append(SpeciesAtIndex)
            links.new(GroupInput.outputs["Geometry"], SpeciesAtIndex.inputs["Geometry"])
            output_socket = get_socket_by_identifier(
                AtomsIndexAttribute, "Attribute_Int", type="outputs"
            )
            links.new(output_socket, SpeciesAtIndex.inputs["Index"])
            input_socket = get_socket_by_identifier(SpeciesAtIndex, "Value_Int")
            output_socket = get_socket_by_identifier(
                SpeciesIndexAttribute, "Attribute_Int", type="outputs"
            )
            links.new(output_socket, input_socket)
        logger.debug("Build geometry nodes for bonds: %s" % (time() - tstart))
        #

    def build_bond_mesh(self):
        """Create geometry node for Bond Mesh.
        1) Transform the edge to mesh points.
        2) Set the position of the points to the center of the bond.
        3) Calculate the bond vector and length.
        """
        from batoms.utils.butils import (
            get_socket_by_identifier,
            get_node_by_name,
            create_node_tree,
        )

        default_interface = [
            ["Geometry", "NodeSocketGeometry", "INPUT"],
            ["Geometry", "NodeSocketGeometry", "OUTPUT"],
            ["Length", "NodeSocketFloat", "OUTPUT"],
            ["Rotation", "NodeSocketVector", "OUTPUT"],
        ]
        parent_tree = self.bond_node.node_tree
        name = "Bond_Mesh_%s" % (self.label)
        node = get_node_by_name(parent_tree.nodes, name=name, type="GeometryNodeGroup")
        node_tree = create_node_tree(name=name, interface=default_interface)
        node.node_tree = node_tree
        nodes = node_tree.nodes
        links = node_tree.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # link the input to parent node
        parent_tree.links.new(
            parent_tree.nodes[0].outputs["Geometry"], node.inputs["Geometry"]
        )
        # -------------------------------------------------------------
        # calculate bond vector, length, rotation based on the index
        # Get four positions from batoms
        # atoms_index0, atoms_index1 of bond
        # atoms_index2, atoms_index3 of the second bond, for high order bond plane
        PositionBatoms = get_node_by_name(
            nodes, "%s_PositionBatoms" % (self.label), "GeometryNodeInputPosition"
        )
        PositionsAtIndexs = []
        AtomsIndexAttributes = []
        for i in range(4):
            tmp = get_node_by_name(
                nodes,
                "%s_PositionsAtIndex%s" % (self.label, i),
                "GeometryNodeSampleIndex",
            )
            tmp.data_type = "FLOAT_VECTOR"
            links.new(GroupInput.outputs["Geometry"], tmp.inputs["Geometry"])
            PositionsAtIndexs.append(tmp)
            tmp = get_node_by_name(
                nodes,
                "%s_NamedAttribute_atoms_index%s" % (self.label, i),
                "GeometryNodeInputNamedAttribute",
            )
            tmp.inputs["Name"].default_value = f"atoms_index{i}"
            tmp.data_type = "INT"
            AtomsIndexAttributes.append(tmp)
        for i in range(4):
            socket = get_socket_by_identifier(PositionsAtIndexs[i], "Value_Vector")
            links.new(PositionBatoms.outputs[0], socket)
            socket = get_socket_by_identifier(
                AtomsIndexAttributes[i], "Attribute_Int", type="outputs"
            )
            links.new(socket, PositionsAtIndexs[i].inputs["Index"])
        # ------------------------------------------------------------------
        # add positions with offsets
        # first step: get offsets from named attribute
        OffsetsAttributes = []
        for i in range(4):
            tmp = get_node_by_name(
                nodes,
                "%s_NamedAttribute_offsets%s" % (self.label, i),
                "GeometryNodeInputNamedAttribute",
            )
            tmp.inputs["Name"].default_value = f"bond_offset{i}"
            tmp.data_type = "FLOAT_VECTOR"
            OffsetsAttributes.append(tmp)
        # second step: we need five add operations
        # four: Get the positions with offset for four atoms
        # one: Get center = (positions1 + positions2)/2
        VectorAdd = []
        for i in range(5):
            tmp = get_node_by_name(
                nodes, "%s_VectorAdd%s" % (self.label, i), "ShaderNodeVectorMath"
            )
            tmp.operation = "ADD"
            VectorAdd.append(tmp)
        for i in range(4):
            socket = get_socket_by_identifier(
                PositionsAtIndexs[i], "Value_Vector", type="outputs"
            )
            links.new(socket, VectorAdd[i].inputs[0])
            links.new(OffsetsAttributes[i].outputs[0], VectorAdd[i].inputs[1])
        # divide by 2 to get the center
        VectorDivide = get_node_by_name(
            nodes, "VectorDivide_%s" % self.label, "ShaderNodeVectorMath"
        )
        VectorDivide.operation = "DIVIDE"
        VectorDivide.inputs[1].default_value = (2, 2, 2)
        links.new(VectorAdd[0].outputs[0], VectorAdd[4].inputs[0])
        links.new(VectorAdd[1].outputs[0], VectorAdd[4].inputs[1])
        links.new(VectorAdd[4].outputs[0], VectorDivide.inputs[0])
        # set center of the bond
        SetPosition = get_node_by_name(
            nodes, "%s_SetPosition" % self.label, "GeometryNodeSetPosition"
        )
        MeshToPoints = get_node_by_name(
            nodes, "%s_MeshToPoints_edge" % (self.label), "GeometryNodeMeshToPoints"
        )
        MeshToPoints.mode = "EDGES"
        links.new(GroupInput.outputs["Geometry"], MeshToPoints.inputs["Mesh"])
        links.new(MeshToPoints.outputs["Points"], SetPosition.inputs["Geometry"])
        links.new(VectorDivide.outputs[0], SetPosition.inputs["Position"])
        links.new(SetPosition.outputs["Geometry"], GroupOutput.inputs["Geometry"])
        # get the vector for the bond and the length
        # also the vector for the second bond
        VectorSubtract = []
        for i in range(2):
            tmp = get_node_by_name(
                nodes, "%s_VectorSubtract%s" % (self.label, i), "ShaderNodeVectorMath"
            )
            tmp.operation = "SUBTRACT"
            VectorSubtract.append(tmp)
        VectorLength = get_node_by_name(
            nodes, "%s_VectorLength" % self.label, "ShaderNodeVectorMath"
        )
        VectorLength.operation = "LENGTH"
        VectorCross0 = get_node_by_name(
            nodes, "%s_VectorCross0" % self.label, "ShaderNodeVectorMath"
        )
        VectorCross0.operation = "CROSS_PRODUCT"
        links.new(VectorAdd[0].outputs[0], VectorSubtract[0].inputs[0])
        links.new(VectorAdd[1].outputs[0], VectorSubtract[0].inputs[1])
        links.new(VectorAdd[2].outputs[0], VectorSubtract[1].inputs[0])
        links.new(VectorAdd[3].outputs[0], VectorSubtract[1].inputs[1])
        # calc the bond length, use it to build scale
        links.new(VectorSubtract[0].outputs[0], VectorLength.inputs[0])
        # cross for rotation, for high order bond
        links.new(VectorSubtract[0].outputs[0], VectorCross0.inputs[0])
        links.new(VectorSubtract[1].outputs[0], VectorCross0.inputs[1])
        # get Euler for rotation
        # we need align two vectors to fix a plane
        # we build the instancer by fix bond diection to Z, and
        # high order bond shift to X
        # thus the the normal of high order bond plane is Y
        AlignEuler = []
        for i in range(2):
            tmp = get_node_by_name(
                nodes,
                "%s_AlignEuler%s" % (self.label, i),
                "FunctionNodeAlignEulerToVector",
            )
            AlignEuler.append(tmp)
        AlignEuler[0].axis = "Z"
        AlignEuler[1].axis = "Y"
        # We should fix Z when align Y
        AlignEuler[1].pivot_axis = "Z"
        links.new(VectorSubtract[0].outputs[0], AlignEuler[0].inputs["Vector"])
        links.new(AlignEuler[0].outputs[0], AlignEuler[1].inputs["Rotation"])
        links.new(VectorCross0.outputs[0], AlignEuler[1].inputs["Vector"])
        links.new(VectorLength.outputs["Value"], GroupOutput.inputs["Length"])
        links.new(AlignEuler[1].outputs[0], GroupOutput.inputs["Rotation"])

    def update_geometry_nodes(self):
        tstart = time()
        # for sp in self.settings:
        # self.add_bond_pair_node(sp.as_dict())
        # only build pair in bondlists
        bondlists = self.arrays
        if len(bondlists["atoms_index0"]) == 0:
            return
        pairs = np.concatenate(
            (
                bondlists["species_index0"].reshape(-1, 1),
                bondlists["species_index1"].reshape(-1, 1),
                bondlists["bond_order"].reshape(-1, 1),
                bondlists["bond_style"].reshape(-1, 1),
            ),
            axis=1,
        )
        pairs = np.unique(pairs, axis=0)
        pairs = pairs.reshape(-1, 4)
        for pair in pairs:
            sp1 = number2String(pair[0])
            sp2 = number2String(pair[1])
            order = pair[2]
            style = pair[3]
            sp = self.settings["%s-%s" % (sp1, sp2)]
            self.add_bond_pair_node(sp.as_dict(), order, style)

        logger.debug("Update geometry nodes for bonds: %s" % (time() - tstart))

    def add_bond_pair_node(self, sp, order=None, style=None):
        """Add geometry node for the bond pair."""
        from batoms.utils.butils import get_node_by_name, get_socket_by_identifier

        parent_tree = self.bond_node.node_tree
        # tstart = time()
        if not order:
            order = sp["order"]
        if not style:
            style = int(sp["style"])
        name = "Bond_%s_%s_%s_%s" % (self.label, sp["name"], order, style)
        node = self.get_bond_pair_node(name)
        nodes = node.node_tree.nodes
        links = node.node_tree.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # link the input to parent node
        BondMesh = get_node_by_name(
            parent_tree.nodes, "Bond_Mesh_%s" % (self.label), "GeometryNodeSetPosition"
        )
        CompareSpecies = []
        for i in range(2):
            SpeciesAtIndex = get_node_by_name(
                parent_tree.nodes,
                "%s_SpeciesAtIndex%s" % (self.label, i),
                "GeometryNodeSampleIndex",
            )
            output_socket = get_socket_by_identifier(
                SpeciesAtIndex, "Value_Int", type="outputs"
            )
            parent_tree.links.new(output_socket, node.inputs[f"Species_index{i}"])
            # e.g C-H bond, species0 is C, species1 is H
            tmp = get_node_by_name(
                nodes,
                "%s_CompareSpecies_%s_%s" % (self.label, sp["species1"], i),
                "FunctionNodeCompare",
            )
            tmp.operation = "EQUAL"
            tmp.data_type = "INT"
            node.inputs[f"Species{i}"].default_value = string2Number(
                sp[f"species{i + 1}"]
            )
            CompareSpecies.append(tmp)
            A_socket = get_socket_by_identifier(tmp, "A_INT")
            B_socket = get_socket_by_identifier(tmp, "B_INT")
            links.new(GroupInput.outputs[f"Species_index{i}"], A_socket)
            links.new(GroupInput.outputs[f"Species{i}"], B_socket)
        node.inputs["Order"].default_value = order
        node.inputs["Style"].default_value = style
        parent_tree.links.new(BondMesh.outputs["Geometry"], node.inputs["Geometry"])
        parent_tree.links.new(BondMesh.outputs["Length"], node.inputs["Length"])
        parent_tree.links.new(BondMesh.outputs["Rotation"], node.inputs["Rotation"])
        # link the outputs to parent node
        JoinGeometry = get_node_by_name(
            parent_tree.nodes,
            "%s_JoinGeometry" % self.label,
            "GeometryNodeJoinGeometry",
        )
        parent_tree.links.new(node.outputs["Geometry"], JoinGeometry.inputs["Geometry"])
        #
        InstanceOnPoint = get_node_by_name(
            nodes, "InstanceOnPoint_%s" % name, "GeometryNodeInstanceOnPoints"
        )
        ObjectInstancer = get_node_by_name(
            nodes, "ObjectInfo_%s" % name, "GeometryNodeObjectInfo"
        )
        ObjectInstancer.inputs["Object"].default_value = self.settings.instancers[
            sp["name"]
        ]["%s_%s" % (order, style)]
        #
        # order
        BondOrderAttribute = get_node_by_name(
            nodes,
            "%s_NamedAttribute_bond_order" % (self.label),
            "GeometryNodeInputNamedAttribute",
        )
        BondOrderAttribute.inputs["Name"].default_value = "bond_order"
        BondOrderAttribute.data_type = "INT"
        CompareOrder = get_node_by_name(
            nodes, "CompareOrder_%s_%s" % (self.label, order), "FunctionNodeCompare"
        )
        CompareOrder.operation = "EQUAL"
        CompareOrder.data_type = "INT"
        socket = get_socket_by_identifier(CompareOrder, "B_INT")
        links.new(GroupInput.outputs["Order"], socket)
        CompareOrder.inputs[1].default_value = order
        output_socket = get_socket_by_identifier(
            BondOrderAttribute, "Attribute_Int", type="outputs"
        )
        input_socket = get_socket_by_identifier(CompareOrder, "A_INT")
        links.new(output_socket, input_socket)
        # style
        BondStyleAttribute = get_node_by_name(
            nodes,
            "%s_NamedAttribute_bond_style" % (self.label),
            "GeometryNodeInputNamedAttribute",
        )
        BondStyleAttribute.inputs["Name"].default_value = "bond_style"
        BondStyleAttribute.data_type = "INT"
        CompareStyle = get_node_by_name(
            nodes, "CompareStyle_%s_%s" % (self.label, style), "FunctionNodeCompare"
        )
        CompareStyle.operation = "EQUAL"
        CompareStyle.data_type = "INT"
        socket = get_socket_by_identifier(CompareStyle, "B_INT")
        links.new(GroupInput.outputs["Style"], socket)
        CompareStyle.inputs[1].default_value = style
        output_socket = get_socket_by_identifier(
            BondStyleAttribute, "Attribute_Int", type="outputs"
        )
        input_socket = get_socket_by_identifier(CompareStyle, "A_INT")
        links.new(output_socket, input_socket)
        BoolSpecies = get_node_by_name(
            nodes, "%s_BooleanMath_species" % name, "FunctionNodeBooleanMath"
        )
        BoolOrder = get_node_by_name(
            nodes, "%s_BooleanMath_order" % name, "FunctionNodeBooleanMath"
        )
        BoolStyle = get_node_by_name(
            nodes, "%s_BooleanMath_style" % name, "FunctionNodeBooleanMath"
        )
        BoolModelStyle = get_node_by_name(
            nodes, "%s_BooleanMath_modelstyle" % name, "FunctionNodeBooleanMath"
        )
        BoolShow = get_node_by_name(
            nodes, "%s_BooleanMath_show" % name, "FunctionNodeBooleanMath"
        )
        BoolBondLength = get_node_by_name(
            nodes, "%s_BoolBondLength" % name, "FunctionNodeBooleanMath"
        )
        # bondlength larger than max will not show
        LessBondLength = get_node_by_name(
            nodes, "%s_LessBondLength" % name, "ShaderNodeMath"
        )
        LessBondLength.operation = "LESS_THAN"
        LessBondLength.inputs[1].default_value = sp["max"]
        #
        links.new(GroupInput.outputs["Geometry"], InstanceOnPoint.inputs["Points"])
        # scale
        CombineXYZ = get_node_by_name(
            nodes, "%s_CombineXYZ" % self.label, "ShaderNodeCombineXYZ"
        )
        CombineXYZ.inputs[0].default_value = 1
        CombineXYZ.inputs[1].default_value = 1
        links.new(GroupInput.outputs["Length"], CombineXYZ.inputs["Z"])
        # show
        BondShowAttribute = get_node_by_name(
            nodes,
            "%s_NamedAttribute_bond_show" % (self.label),
            "GeometryNodeInputNamedAttribute",
        )
        BondShowAttribute.inputs["Name"].default_value = "bond_show"
        BondShowAttribute.data_type = "INT"
        socket = get_socket_by_identifier(
            BondShowAttribute, "Attribute_Int", type="outputs"
        )
        links.new(socket, BoolShow.inputs[0])
        # model style
        BondModelStyleAttribute = get_node_by_name(
            nodes,
            "%s_NamedAttribute_bond_model_style" % (self.label),
            "GeometryNodeInputNamedAttribute",
        )
        BondModelStyleAttribute.inputs["Name"].default_value = "bond_model_style"
        BondModelStyleAttribute.data_type = "INT"
        socket = get_socket_by_identifier(
            BondModelStyleAttribute, "Attribute_Int", type="outputs"
        )
        links.new(socket, BoolModelStyle.inputs[0])
        for i in range(2):
            links.new(CompareSpecies[i].outputs[0], BoolSpecies.inputs[i])
        links.new(BoolSpecies.outputs[0], BoolOrder.inputs[0])
        links.new(CompareOrder.outputs[0], BoolOrder.inputs[1])
        links.new(BoolOrder.outputs[0], BoolStyle.inputs[0])
        links.new(CompareStyle.outputs[0], BoolStyle.inputs[1])
        links.new(BoolStyle.outputs[0], BoolModelStyle.inputs[1])
        links.new(BoolModelStyle.outputs[0], BoolShow.inputs[1])
        links.new(GroupInput.outputs["Length"], LessBondLength.inputs[0])
        links.new(BoolShow.outputs["Boolean"], BoolBondLength.inputs[0])
        links.new(LessBondLength.outputs[0], BoolBondLength.inputs[1])
        links.new(
            BoolBondLength.outputs["Boolean"], InstanceOnPoint.inputs["Selection"]
        )
        links.new(CombineXYZ.outputs[0], InstanceOnPoint.inputs["Scale"])
        #
        links.new(GroupInput.outputs["Rotation"], InstanceOnPoint.inputs["Rotation"])
        #
        links.new(
            ObjectInstancer.outputs["Geometry"], InstanceOnPoint.inputs["Instance"]
        )
        links.new(InstanceOnPoint.outputs["Instances"], GroupOutput.inputs["Geometry"])
        # print('Add geometry nodes for bonds: %s'%(time() - tstart))

    def update_geometry_node_instancer(self):
        """
        Make sure all pair has a geometry node flow
        and the instancer and material are updated.
        """
        tstart = time()
        #
        bondlists = self.arrays
        if len(bondlists["atoms_index0"]) == 0:
            return
        pairs = np.concatenate(
            (
                bondlists["species_index0"].reshape(-1, 1),
                bondlists["species_index1"].reshape(-1, 1),
                bondlists["bond_order"].reshape(-1, 1),
                bondlists["bond_style"].reshape(-1, 1),
            ),
            axis=1,
        )
        pairs = np.unique(pairs, axis=0)
        pairs = pairs.reshape(-1, 4)
        for pair in pairs:
            sp1 = number2String(pair[0])
            sp2 = number2String(pair[1])
            order = pair[2]
            style = pair[3]
            order_style = "%s_%s" % (order, style)
            sp = self.settings["%s-%s" % (sp1, sp2)]
            sp = sp.as_dict()
            # update geometry node
            if self.settings.instancers[sp["name"]][order_style] is None:
                self.settings.build_instancer(sp, order, style)
                self.add_bond_pair_node(sp)
            # always update materials
            self.settings.build_materials(sp, material_style=sp["material_style"])
            self.settings.assign_materials(sp, sp["order"], sp["style"])
            # compare radius
            if not np.isclose(
                sp["width"],
                self.settings.instancers[sp["name"]][order_style].Bbond.width,
            ):
                self.settings.build_instancer(sp)
                self.settings.assign_materials(sp, sp["order"], sp["style"])
            # update  instancers
            name = "Bond_%s_%s_%s_%s" % (
                self.label,
                sp["name"],
                sp["order"],
                sp["style"],
            )
            node = self.get_bond_pair_node(name)
            ObjectInstancer = get_node_by_name(
                node.node_tree.nodes, "ObjectInfo_%s" % name, "GeometryNodeObjectInfo"
            )
            ObjectInstancer.inputs["Object"].default_value = self.settings.instancers[
                sp["name"]
            ][order_style]
        logger.debug("update bond instancer: %s" % (time() - tstart))

    def update(self, bondlists=None, orders=None):
        """
        Draw bonds.
        calculate bond in all farmes, and merge all bondlists.
        Draw bonds in the bondlists, only show the bond if it
        is smaller than the max bond length.
        """

        object_mode()
        # clean_coll_objects(self.coll, 'bond')
        positions = self.batoms.positions
        trajectory = self.batoms.get_trajectory()["positions"]
        arrays = self.batoms.arrays
        boundary_data = self.batoms.boundary.boundary_data
        show = arrays["show"].astype(bool)
        species = arrays["species"][show]
        # trajectory_boundary = self.batoms.get_trajectory(self.batoms.batoms_boundary)
        # trajectory_search = self.batoms.get_trajectory(self.batoms.batoms_search)
        if self.batoms.nframe == 0:
            trajectory = np.array([positions])
        nframe = len(trajectory)
        bond_datas = {}
        tstart = time()
        setting = self.settings.as_dict()
        if bondlists is None:
            for f in range(nframe):
                # print('update bond: ', f)
                positions = trajectory[f, show, :]
                # build bondlist for unit cell
                (
                    bondlist,
                    bonddatas,
                    peciesBondDatas,
                    molPeciesDatas,
                ) = self.build_bondlists(
                    species, positions, self.batoms.cell, self.batoms.pbc, setting
                )
                # build bondlist for boundary atoms
                # for molecule with cell == [0, 0, 0], skip
                if boundary_data is not None:
                    bondlist = self.build_bondlists_with_boundary(
                        boundary_data,
                        bondlist,
                        bonddatas,
                        peciesBondDatas,
                        molPeciesDatas,
                    )
                    bondlist = self.check_boundary(bondlist)
                # search molecule
                # search bond
                if self.show_search:
                    self.search_bond.hide = False
                    self.search_bond.update(
                        bondlist,
                        peciesBondDatas,
                        molPeciesDatas,
                        arrays,
                        self.batoms.cell,
                    )
                if f == 0:
                    bondlists = bondlist
                else:
                    bondlists = np.append(bondlists, bondlist, axis=0)
            bondlists = np.unique(bondlists, axis=0)
        bond_datas = self.calc_bond_data(
            species,
            trajectory[:, show, :],
            self.batoms.cell,
            bondlists,
            self.settings,
            arrays["model_style"][show],
        )
        if orders:
            bond_datas.update({"bond_order": orders})
        if len(bond_datas) == 0:
            return
        self.set_arrays(bond_datas)
        logger.debug("draw bond: {0:10.2f} s".format(time() - tstart))

    def get_arrays(self):
        """ """
        object_mode()
        # tstart = time()
        arrays = self.get_attributes(domains=["EDGE"])
        arrays.update(
            {
                "positions": self.positions,
            }
        )
        # print('get_arrays: %s'%(time() - tstart))
        return arrays

    def set_arrays(self, arrays):
        """Add edge for bond.
        Set the bond properties."""
        # if not attributes, set them
        # in the case of mergeing objects, e.g. au111 += h2o,
        # the attributes of edges are removed.
        if "atoms_index0" not in self.obj.data.attributes:
            for att in default_bond_attributes:
                self.add_attribute(**att)
        # blender delete all edges
        dnvert = len(self.obj.data.edges)
        self.delete_edges_bmesh(range(dnvert))
        #
        edges = np.vstack((arrays["atoms_index0"], arrays["atoms_index1"])).T
        self.add_edges(edges, self.batoms.obj)
        self.set_attributes(arrays)
        # self.set_trajectory(arrays)
        self.update_geometry_node_instancer()
        self.update_geometry_nodes()

    def get_trajectory(self):
        """ """
        trajectory = {}
        trajectory["centers"] = self.get_obj_trajectory(self.obj)
        # trajectory['offsets'] = self.get_obj_trajectory(self.obj_o)
        return trajectory

    def set_trajectory(self, trajectory=None, frame_start=0):
        if trajectory is None:
            trajectory = self._trajectory
        nframe = len(trajectory["centers"])
        if nframe == 0:
            return
        name = "%s_bond" % (self.label)
        obj = self.obj
        self.set_shape_key(name, obj, trajectory["centers"], frame_start=frame_start)
        #
        # name = '%s_bond_offset'%(self.label)
        # obj = self.obj_o
        # self.set_shape_key(name, obj, trajectory['offsets'], frame_start=frame_start)

    @property
    def bondlists(self):
        return self.get_bondlists()

    def get_bondlists(self):
        """ """
        object_mode()
        # tstart = time()
        arrays = self.arrays
        i = arrays["atoms_index0"].reshape(-1, 1)
        j = arrays["atoms_index1"].reshape(-1, 1)
        p = arrays["polyhedra"].reshape(-1, 1)
        bond_offset0 = arrays["bond_offset0"]
        bond_offset1 = arrays["bond_offset1"]
        bondlists = np.concatenate((i, j, bond_offset0, bond_offset1, p), axis=1)
        # bondlists = bondlists.astype(int)
        # print('get_arrays: %s'%(time() - tstart))
        return bondlists

    @property
    def search_bond(self):
        """search_bond object."""
        if self._search_bond is not None:
            return self._search_bond
        search_bond = SearchBond(self.label, batoms=self.batoms)
        self._search_bond = search_bond
        return search_bond

    @property
    def show_search(self):
        return self.settings.coll.Bbond.show_search

    @show_search.setter
    def show_search(self, show_search):
        self.settings.coll.Bbond.show_search = show_search
        if not show_search:
            self.search_bond.set_arrays(default_search_bond_datas.copy())
            self._search_bond = None
        else:
            self.update()

    @property
    def show_hydrogen_bond(self):
        return self.settings.coll.Bbond.show_hydrogen_bond

    @show_hydrogen_bond.setter
    def show_hydrogen_bond(self, show_hydrogen_bond):
        self.settings.coll.Bbond.show_hydrogen_bond = show_hydrogen_bond
        self.update()

    def __getitem__(self, indices):
        """Return a subset of the Bbond.

        i -- int, describing which atom to return.

        #todo: this is slow for large system

        """
        from batoms.bond.slicebonds import SliceBonds

        slicebonds = SliceBonds(self.label, indices, bonds=self)
        return slicebonds

    def __setitem__(self, indices, value):
        """Return a subset of the Bbond.

        i -- int, describing which atom to return.

        #todo: this is slow for large system

        """
        positions = self.positions
        positions[indices] = value
        self.set_positions(positions)

    def repeat(self, m, cell):
        """
        In-place repeat of atoms.

        >>> from batoms.bond import Bbond
        >>> c = Bbond('co', 'C', [[0, 0, 0], [1.2, 0, 0]])
        >>> c.repeat([3, 3, 3], np.array([[5, 0, 0], [0, 5, 0], [0, 0, 5]]))
        """
        if isinstance(m, int):
            m = (m, m, m)
        for x, vec in zip(m, cell):
            if x != 1 and not vec.any():
                raise ValueError("Cannot repeat along undefined lattice " "vector")
        M = np.product(m)
        n = len(self)
        positions = np.tile(
            self.positions, (M,) + (1,) * (len(self.positions.shape) - 1)
        )
        i0 = 0
        for m0 in range(m[0]):
            for m1 in range(m[1]):
                for m2 in range(m[2]):
                    i1 = i0 + n
                    positions[i0:i1] += np.dot((m0, m1, m2), cell)
                    i0 = i1
        self.add_vertices(positions[n:])

    def copy(self, label, species):
        """
        Return a copy.

        name: str
            The name of the copy.

        For example, copy H species:

        >>> h_new = h.copy(label = 'h_new', species = 'H')

        """
        object_mode()
        bbond = self.__class__(
            label,
            species,
            self.local_positions,
            location=self.obj.location,
            scale=self.scale,
            material=self.material,
        )
        return bbond

    def extend(self, other):
        """
        Extend bbond object by appending bbond from *other*.

        >>> from batoms.bonds import Bbond
        >>> h1 = Bbond('h2o', 'H_1', [[0, 0, 0], [2, 0, 0]])
        >>> h2 = Bbond('h2o', 'H_2', [[0, 0, 2], [2, 0, 2]])
        >>> h = h1 + h2
        """
        # could also use self.add_vertices(other.positions)
        object_mode()
        bpy.ops.object.select_all(action="DESELECT")
        self.obj.select_set(True)
        other.obj.select_set(True)
        bpy.context.view_layer.objects.active = self.obj
        bpy.ops.object.join()

    def __iadd__(self, other):
        """
        >>> h1 += h2
        """
        self.extend(other)
        return self

    def __add__(self, other):
        """
        >>> h1 = h1 + h2
        """
        self += other
        return self

    def __iter__(self):
        bbond = self.obj
        for i in range(len(self)):
            yield bbond.matrix_world @ bbond.data.vertices[i].co

    def __repr__(self):
        s = "Bonds(Total: {:6d}, {}".format(len(self), self.arrays)
        return s

    def build_bondlists(self, species, positions, cell, pbc, setting):
        """
        build bondlist for atoms
        steps:
        1 build bondlist for atoms with pbc
        2 search connected_components (molecule),
          return pecies data and its neighbour
        3 add bondlist related with molecule

        """
        from batoms.neighborlist import bondlist_kdtree

        bondlists = np.zeros((0, 11), dtype=int)
        bonddatas = {}
        if len(setting) == 0:
            return bondlists, bonddatas, {}, {}
        #
        # tstart = time()
        # ==========================================================
        # step 1 build bondlist for atoms with pbc
        # nli: index1
        # nlj: index2
        # nlk: search bond style
        # nlp: polyhedra
        # nlt: bond type: hydrogen bond
        # nlSj: offset of atoms in nlj
        nli, nlj, nlk, nlp, nlt, nlSj = bondlist_kdtree(
            "ijkptS", species, positions, cell, pbc, setting
        )
        nb = len(nli)
        nlSi = np.zeros((nb, 3))
        # print('build_bondlists: {0:10.2f} s'.format(time() - tstart))
        # TODO search0 is not used, please check why
        # search type 0,
        # search0 = np.where(
        # (nlk == 0) & (nlSj != np.array([0, 0, 0])).any(axis=1), False, True
        # )
        # 0  1  2:5       5:8          8     9
        # i, j, offset_i, offset_j, search, search_style
        bondlists = np.concatenate(
            (
                np.array([nli, nlj]).T,
                np.array(nlSi, dtype=int),
                np.array(nlSj),
                np.array(nlk).reshape(-1, 1),
                np.array(nlp).reshape(-1, 1),
                np.array(nlt).reshape(-1, 1),
            ),
            axis=1,
        )
        bondlists = bondlists.astype(int)
        # remove bond outside box for search0
        # not now, we need all bonds here, and then add a final check.
        # bondlists = bondlists[search0]
        # build bondatas, for each atom, save the bonds connect to it.
        argsort = bondlists[:, 0].argsort()
        bondlists = bondlists[argsort]
        u, indices = np.unique(bondlists[:, 0], return_index=True)
        indices = np.append(indices, len(bondlists))
        m = len(u)
        for i in range(m):
            bonddatas[u[i]] = bondlists[indices[i] : indices[i + 1]]
        #
        # ===================================================
        # 2 search connected_components (molecule),
        #   return pecies data and its neighbour
        peciesBondLists, molPeciesDatas = self.build_peciesBondLists(
            len(positions), bondlists
        )
        # build peciesBondDatas
        peciesBondDatas = {}
        for p in molPeciesDatas:
            peciesBondDatas[p] = []
        # since this is search bond type 2, bothway, first i
        argsort = peciesBondLists[:, 0].argsort()
        peciesBondLists = peciesBondLists[argsort]
        u, indices = np.unique(peciesBondLists[:, 0], return_index=True)
        indices = np.append(indices, len(peciesBondLists))
        m = len(u)
        # for each pecies, save its neighbour and offset
        for i in range(m):
            peciesBondDatas[u[i]] = peciesBondLists[indices[i] : indices[i + 1]]
        # bothway, then j
        argsort = peciesBondLists[:, 1].argsort()
        peciesBondLists = peciesBondLists[argsort]
        u, indices = np.unique(peciesBondLists[:, 1], return_index=True)
        indices = np.append(indices, len(peciesBondLists))
        m = len(u)
        # for each pecies, save its neighbour and offset
        for i in range(m):
            data = peciesBondLists[indices[i] : indices[i + 1]]
            data = data.reshape(-1, 11)
            data[:, 5:8] *= -1
            data[:, [0, 1]] = data[:, [1, 0]]
            peciesBondDatas[u[i]] = data
        # ==========================================================
        # 3 add bondlist related with molecule
        # add bondlist for molecule
        for mol in peciesBondLists:
            indices = molPeciesDatas[mol[1]]
            for i in indices:
                if i not in bonddatas:
                    continue
                data = bonddatas[i]
                n = len(data)
                bondlists = np.append(bondlists, data, axis=0)
                bondlists[-n:, 2:5] += mol[5:8]
                bondlists[-n:, 5:8] += mol[5:8]
            indices = molPeciesDatas[mol[0]]
            for i in indices:
                if i not in bonddatas:
                    continue
                data = bonddatas[i]
                n = len(data)
                bondlists = np.append(bondlists, data, axis=0)
                bondlists[-n:, 2:5] -= mol[5:8]
                bondlists[-n:, 5:8] -= mol[5:8]
        # print(bondlists)
        bondlists = bondlists.astype(int)
        bondlists = np.unique(bondlists, axis=0)
        self.peciesBondLists = peciesBondLists
        self.molPeciesDatas = molPeciesDatas
        self.peciesBondDatas = peciesBondDatas
        return bondlists, bonddatas, peciesBondDatas, molPeciesDatas

    def build_peciesBondLists(self, natom, bondlists):
        """
        search type 2: build molecule
        steps:
        1 search connected_components (molecules) inside atoms
        2 construct the molecules by its pecies and the coresponding offsets
        3 for
        3 return the molecules
        """
        from scipy.sparse import csgraph, csr_matrix

        molPeciesDatas = {}
        peciesBondLists = np.zeros((0, 11), dtype=int)
        # search type 2
        k = bondlists[:, 8]
        indices = np.where(k == 2)[0]
        ns2 = len(indices)
        if ns2 == 0:
            return peciesBondLists, molPeciesDatas
        bondlists1 = bondlists[indices, :]
        # ========================================================
        # 1 search connected_components (molecules) inside atoms
        # with crossed bond, entire molecule
        molDatas = {}
        ai = bondlists1[:, 0]
        aj = bondlists1[:, 1]
        data = np.ones(ns2, dtype=int)
        matrix = csr_matrix((data, (ai, aj)), shape=(natom, natom))
        n_components1, component_list1 = csgraph.connected_components(matrix)
        for i in range(n_components1):
            indices = np.where(component_list1 == i)[0]
            n = len(indices)
            if n < 2:
                continue
            molDatas[i] = {"sub": []}
            molDatas[i]["indices"] = indices
            # TODO: check here
            molDatas[i]["offsets"] = indices
        # without crossed bond, small pecies of molDatas
        mask = np.where(
            (bondlists1[:, 2:5] != bondlists1[:, 5:8]).any(axis=1), False, True
        )
        data = data[mask]
        ai = ai[mask]
        aj = aj[mask]
        matrix = csr_matrix((data, (ai, aj)), shape=(natom, natom))
        n_components2, component_list2 = csgraph.connected_components(matrix)
        #
        for i in range(n_components2):
            indices = np.where(component_list2 == i)[0]
            if component_list1[indices[0]] in molDatas:
                # this pecies belong to molDatas
                molDatas[component_list1[indices[0]]]["sub"].append(i)
                molPeciesDatas[i] = indices
        # pprint(molPeciesDatas)
        # cross box bond, find direct neighbour pecies
        bondlists2 = bondlists1[~mask]
        n = len(bondlists2)
        ai = bondlists2[:, 0].astype(int)
        aj = bondlists2[:, 1].astype(int)
        peciesBondLists = np.zeros((n, 11), dtype=int)
        molBondDicts = {}
        for i in range(n):
            # molDatasId = component_list1[bondlists2[i, 0]]
            i1 = component_list2[ai[i]]
            j1 = component_list2[aj[i]]
            peciesBondLists[i, 0] = i1
            peciesBondLists[i, 1] = j1
            peciesBondLists[i, 2:] = bondlists2[i, 2:]
            molBondDicts[(i1, j1)] = bondlists2[i, 2:]
        peciesBondLists = np.unique(peciesBondLists, axis=0)
        # ========================================================
        # find indirect neighbour pecies inside mol1
        # find connected path between pecies
        ai = peciesBondLists[:, 0].astype(int)
        aj = peciesBondLists[:, 1].astype(int)
        nml = len(ai)
        data = np.ones(nml, dtype=int)
        graph = csr_matrix((data, (ai, aj)), shape=(n_components2, n_components2))
        dist_matrix, predecessors = csgraph.shortest_path(
            csgraph=graph, return_predecessors=True
        )
        # print('dist_matrix: ', dist_matrix)
        # print('predecessors: ', predecessors)
        dist_matrix = dist_matrix.astype(int)
        mollist = np.zeros(11, dtype=int)
        for i, data in molDatas.items():
            indices = data["sub"]
            n = len(indices)
            if n < 2:
                continue
            for j in range(n - 1):
                for k in range(j + 1, n):
                    if dist_matrix[indices[j], indices[k]] == 1:
                        continue
                    path = [indices[k]]
                    end = indices[k]
                    last = indices[k]
                    for i1 in range(dist_matrix[indices[j], end]):
                        last = predecessors[indices[j], last]
                        path = [last] + path
                    offsets = np.zeros(3)
                    for i2 in range(0, len(path) - 1):
                        offsets += molBondDicts[(path[i2], path[i2 + 1])][3:6]
                    mollist[0] = path[0]
                    mollist[1] = path[-1]
                    mollist[5:8] = offsets
                    peciesBondLists = np.append(
                        peciesBondLists, np.array([mollist]), axis=0
                    )
        self.molDatas = molDatas
        return peciesBondLists, molPeciesDatas

    def build_bondlists_with_boundary(
        self, arrays, bondlists, bonddatas, peciesBondDatas, molPeciesDatas
    ):
        """
        build extra bondlists based on boundary atoms
        """
        n = len(arrays["positions"])
        if n == 0:
            return bondlists
        # search bond type 0 and 1
        for i in range(n):
            if arrays["indices"][i] not in bonddatas:
                # print('no bonds: ', arrays['indices'][i])
                continue
            data = bonddatas[arrays["indices"][i]]
            n = len(data)
            bondlists = np.append(bondlists, data, axis=0)
            bondlists[-n:, 2:5] = bondlists[-n:, 2:5] + arrays["boundary_offset"][i]
            bondlists[-n:, 5:8] = bondlists[-n:, 5:8] + arrays["boundary_offset"][i]
            # todo: in this case, some of the atoms overlap
            # with original atoms.
            # since it doesn't influence the 3d view, we
            # just leave it like this
            # bondlists[-n:, 8] = 1
        # print('build_bondlists: {0:10.2f} s'.format(time() - tstart))
        bondlists = np.unique(bondlists, axis=0)
        # search bond type 2
        # divide boudanry atoms by boundary_offset
        boundary_offset = np.unique(arrays["boundary_offset"], axis=0)
        n = len(boundary_offset) * len(molPeciesDatas)
        peciesArrays = np.zeros((n, 4), dtype=int)
        n = 0
        for offset in boundary_offset:
            indices = np.where((arrays["boundary_offset"] == offset).all(axis=1))[0]
            indices = arrays["indices"][indices]
            for p, indices0 in molPeciesDatas.items():
                indices1 = np.intersect1d(indices0, indices)
                if len(indices1) > 0:
                    peciesArrays[n][0] = p
                    peciesArrays[n][1:4] = offset
                    n += 1
        peciesArrays = peciesArrays[:n]
        npa = len(peciesArrays)
        if npa == 0:
            return bondlists
        # search bond type 2
        for i in range(npa):
            peciesId = peciesArrays[i, 0]
            p = peciesBondDatas[peciesId]
            # repeat itself
            indices = molPeciesDatas[peciesId]
            for j in indices:
                if j not in bonddatas:
                    continue
                data = bonddatas[j]
                n = len(data)
                bondlists = np.append(bondlists, data, axis=0)
                bondlists[-n:, 2:5] += peciesArrays[i, 1:4]
                bondlists[-n:, 5:8] += peciesArrays[i, 1:4]
            # repeat its neighbour pecies
            for pb in p:
                peciesId = pb[1]
                indices = molPeciesDatas[peciesId]
                for j in indices:
                    if j not in bonddatas:
                        continue
                    data = bonddatas[j]
                    n = len(data)
                    bondlists = np.append(bondlists, data, axis=0)
                    bondlists[-n:, 2:5] += peciesArrays[i, 1:4] + pb[5:8]
                    bondlists[-n:, 5:8] += peciesArrays[i, 1:4] + pb[5:8]
            # todo: in this case, some of the atoms overlap with
            # original atoms.
            # since it doesn't influence the 3d view,
            # we just leave it like this
            # bondlists[-n:, 8] = 1
        # print('build_bondlists: {0:10.2f} s'.format(time() - tstart))
        bondlists = np.unique(bondlists, axis=0)
        # search bond type 2

        return bondlists

    def check_boundary(self, bondlists, eps=1e-6):
        """check boundary for bond search 0

        Args:
            eps (float): default 1e-6
        """
        from ase.geometry import complete_cell

        arrays = self.batoms.arrays
        positions = arrays["positions"]
        cell = self.batoms.cell
        # get scaled positions
        positions = np.linalg.solve(complete_cell(cell).T, positions.T).T
        npositions = positions[bondlists[:, 1]] + bondlists[:, 5:8]
        boundary = self.batoms.boundary.boundary
        # boundary condition
        mask1 = np.where(
            (npositions[:, 0] > boundary[0][0] - eps)
            & (npositions[:, 0] < boundary[0][1] + eps)  # noqa W503
            & (npositions[:, 1] > boundary[1][0] - eps)  # noqa W503
            & (npositions[:, 1] < boundary[1][1] + eps)  # noqa W503
            & (npositions[:, 2] > boundary[2][0] - eps)  # noqa W503
            & (npositions[:, 2] < boundary[2][1] + eps),  # noqa W503
            False,
            True,
        )
        mask2 = np.where(bondlists[:, 8] == 0, True, False)
        mask = ~(mask1 & mask2)
        bondlists = bondlists[mask]
        return bondlists

    def search_molecule(self, natom, bondlists):
        """ """
        from scipy.sparse import csgraph, csr_matrix

        mols = {}
        # search type
        k = bondlists[:, -1]
        indices = np.where(k == 2)[0]
        ns2 = len(indices)
        if ns2 == 0:
            return mols
        i = bondlists[indices, 0]
        j = bondlists[indices, 1]
        data = np.ones(ns2, dtype=int)
        matrix = csr_matrix((data, (i, j)), shape=(natom, natom))
        # print(matrix)
        n_components, component_list = csgraph.connected_components(matrix)
        # print(n_components)
        # print(component_list)
        for i in range(n_components):
            indices = np.where(component_list == i)[0]
            if len(indices) < 2:
                continue
            mols[i] = indices
        return mols

    def calc_bond_data(
        self, speciesarray, positions, cell, bondlists, bondsetting, model_styles
    ):
        """
        positions: 3-demision
        todo: support frame for positions and offsets.
        """
        tstart = time()
        if not self.show_hydrogen_bond:
            bondlists = bondlists[bondlists[:, 10] != 1]
        # properties
        atoms_index0 = np.array(bondlists[:, 0], dtype=int)
        atoms_index1 = np.array(bondlists[:, 1], dtype=int)
        # second bond for high order bond
        nb = len(atoms_index0)
        second_bond = np.roll(np.arange(nb), 1)
        # atoms_index2 and atoms_index3 will be removed for Blender 3.1
        atoms_index2 = np.roll(atoms_index0, 1)
        atoms_index3 = np.roll(atoms_index1, 1)
        shows = np.ones(nb, dtype=int)
        orders = np.zeros(nb, dtype=int)  # 1, 2, 3
        styles = np.zeros(nb, dtype=int)  # 0, 1, 2
        widths = np.ones(nb, dtype=float)
        species_index0 = np.zeros(nb, dtype=int)
        species_index1 = np.zeros(nb, dtype=int)
        bond_model_styles = model_styles[atoms_index0]
        if nb == 0:
            return default_bond_datas
        # ------------------------------------
        # offsets
        bond_offset0 = np.dot(bondlists[:, 2:5], cell)
        bond_offset1 = np.dot(bondlists[:, 5:8], cell)
        # polyhedra
        polyhedras = np.array(bondlists[:, 9], dtype=int)
        # bond center
        if len(positions.shape) == 2:
            positions = np.array([positions])
        # ---------------------------------------------
        for b in bondsetting:
            spi = b.species1
            spj = b.species2
            indi = speciesarray[atoms_index0] == spi
            indj = speciesarray[atoms_index1] == spj
            ind = indi & indj
            orders[ind] = b.order
            styles[ind] = int(b.style)
            widths[ind] = b.width
            polyhedras[ind] = b.polyhedra
            species_index0[ind] = string2Number(spi)
            species_index1[ind] = string2Number(spj)
            if b.order > 1:
                second_bond, atoms_index2, atoms_index3 = self.secondBond(
                    b, speciesarray, second_bond, atoms_index2, atoms_index3, bondlists
                )
        # offsets for second bond
        # this will be remove for Blender 3.1
        bond_offset2 = np.dot(bondlists[:, 2:5][second_bond], cell)
        bond_offset3 = np.dot(bondlists[:, 5:8][second_bond], cell)
        datas = {
            "atoms_index0": atoms_index0,
            "atoms_index1": atoms_index1,
            "atoms_index2": atoms_index2,
            "atoms_index3": atoms_index3,
            "second_bond": second_bond,
            "species_index0": species_index0,
            "species_index1": species_index1,
            "bond_offset0": bond_offset0,
            "bond_offset1": bond_offset1,
            "bond_offset2": bond_offset2,
            "bond_offset3": bond_offset3,
            # 'widths': widths,
            "bond_show": shows,
            "bond_order": orders,
            "bond_style": styles,
            "bond_model_style": bond_model_styles,
            "polyhedra": polyhedras,
        }
        # print('datas: ', datas)
        logger.debug("calc_bond_data: {0:10.2f} s".format(time() - tstart))
        return datas

    def secondBond(
        self, b, speciesarray, second_bond, atoms_index2, atoms_index3, bondlists
    ):
        """
        determine the plane of high order bond.
        v1 = atoms2 - atoms1
        v2 = atoms4 - atoms3
        normal = np.cross(v1, v2)
        """
        spi = b.species1
        spj = b.species2
        # find all bonds belong to this pair spi--spj
        # here use &
        indi = speciesarray[bondlists[:, 0]] == spi
        indj = speciesarray[bondlists[:, 1]] == spj
        indices1 = np.where(indi & indj)[0]
        # find all bonds connect to this bond pair
        # here use |
        indi1 = speciesarray[bondlists[:, 1]] == spi
        indj2 = speciesarray[bondlists[:, 0]] == spj
        indices2 = np.where(indi | indj | indi1 | indj2)[0]
        bondlists2 = bondlists[indices2]
        for i in indices1:
            # find another bond for this bond: bondlists[i]
            indices3 = np.where(
                (bondlists2[:, 0] == bondlists[i, 0])  # noqa W503
                | (bondlists2[:, 1] == bondlists[i, 0])  # noqa W503
                | (bondlists2[:, 0] == bondlists[i, 1])  # noqa W503
                | (bondlists2[:, 1] == bondlists[i, 1])  # noqa W503
            )[0]
            indices3 = indices2[indices3]
            if len(indices3) > 1:
                indices4 = np.where(indices3 != i)[0]
                second_bond[i] = indices3[indices4[0]]
                atoms_index2[i] = bondlists[second_bond[i], 0]
                atoms_index3[i] = bondlists[second_bond[i], 1]
        return second_bond, atoms_index2, atoms_index3

    def high_order_bond_plane(
        self, b, speciesarray, positions, nvec, offsets, bondlists
    ):
        """
        determine the plane of high order bond
        """
        spi = b.species1
        spj = b.species2
        indi = speciesarray[bondlists[:, 0]] == spi
        indj = speciesarray[bondlists[:, 1]] == spj
        bondlists1 = bondlists[indi & indj]
        nbond = len(bondlists1)
        indi1 = speciesarray[bondlists[:, 1]] == spi
        indj2 = speciesarray[bondlists[:, 0]] == spj
        bondlists2 = bondlists[indi | indj | indi1 | indj2]
        for i in range(nbond):
            # find another bond
            mask = np.logical_not(
                (
                    (bondlists2[:, 0] == bondlists1[i, 0])
                    & (bondlists2[:, 1] == bondlists1[i, 1])  # noqa W503
                )
                | (  # noqa W503
                    (bondlists2[:, 0] != bondlists1[i, 0])
                    & (bondlists2[:, 0] != bondlists1[i, 1])  # noqa W503
                    & (bondlists2[:, 1] != bondlists1[i, 0])  # noqa W503
                    & (bondlists2[:, 1] != bondlists1[i, 1])  # noqa W503
                )
            )
            localbondlist = bondlists2[mask]
            if len(localbondlist) == 0:
                second_bond = np.array([0.0, 0.0, 1])
            else:
                second_bond = (
                    positions[localbondlist[0, 0]] - positions[localbondlist[0, 1]]
                )
            norml = np.cross(second_bond, nvec[i]) + np.array([1e-8, 0, 0])
            offset = np.cross(norml, nvec[i]) + np.array([1e-8, 0, 0])
            offsets[i] = offset / np.linalg.norm(offset)
            # print(offsets[i], offset)

        return offsets

    def bond_order_auto_set(self):
        """Set bond order by pybel
        #TODO support show, select
        """
        from openbabel import pybel

        species = self.batoms.arrays["species"]

        mol = self.batoms.as_pybel(export_bond=False)
        bondlists = [
            [
                b.GetBeginAtom().GetIndex(),
                b.GetEndAtom().GetIndex(),
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ]
            for b in pybel.ob.OBMolBondIter(mol.OBMol)
        ]
        # TODO detect hydrogen bond
        cutoff_dict = self.settings.cutoff_dict
        nbond = len(bondlists)
        remove = []
        bondlists = np.array(bondlists)
        for i in range(nbond):
            b = bondlists[i]
            if (species[b[0]], species[b[1]]) not in cutoff_dict:
                # swtich bond
                if (species[b[1]], species[b[0]]) in cutoff_dict:
                    bondlists[i][0], bondlists[i][1] = bondlists[i][1], bondlists[i][0]
                else:
                    # remove bond
                    remove.append(i)
        np.delete(bondlists, remove)
        orders = [b.GetBondOrder() for b in pybel.ob.OBMolBondIter(mol.OBMol)]
        self.update(bondlists, orders=orders)

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {"array": None}
        if len(self) > 1:
            data["array"] = dict(self.arrays)
        data["show_search"] = self.show_search
        data["show_hydrogen_bond"] = self.show_hydrogen_bond
        data["settings"] = self.settings.as_dict()
        return data



from .polyhedra import Polyhedra

from . import (
    bpy_data,
    ops,
    ui_list,
)

__all__ = ["Polyhedra"]

classes = [
    # internal data first
    bpy_data.PolyhedraSetting,
    bpy_data.Polyhedra,
    ops.PolyhedraAdd,
    ops.PolyhedraRemove,
    ops.PolyhedraDraw,
    ops.PolyhedraModify,
    ui_list.BATOMS_MT_polyhedra_context_menu,
    ui_list.BATOMS_UL_polyhedra,
    ui_list.BATOMS_PT_polyhedra,
]


def register_class():
    from bpy.types import Collection, Object
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)
    # attach to blender internal data
    Collection.Bpolyhedra = PointerProperty(name="Bpolyhedra", type=bpy_data.Polyhedra)
    Object.Bpolyhedra = PointerProperty(
        name="Bpolyhedra", type=bpy_data.PolyhedraSetting
    )


def unregister_class():
    from bpy.types import Collection, Object
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del Collection.Bpolyhedra
    del Object.Bpolyhedra



"""
# TODO add locaiton in geometry node
# TODO rearrange code to handle offsets for all bonds, boundary
       search_bond and so on
"""
from time import time

import bpy
import numpy as np
from ase.geometry import complete_cell
from batoms.base.object import ObjectGN
from batoms.utils import number2String, string2Number
from batoms.utils.butils import compareNodeType
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)

default_attributes = [
    {"name": "atoms_index", "data_type": "INT"},
    {"name": "species_index", "data_type": "INT"},
    {"name": "show", "data_type": "INT"},
    {"name": "select", "data_type": "INT"},
    {"name": "model_style", "data_type": "INT"},
    {"name": "scale", "data_type": "FLOAT"},
    {"name": "radius_style", "data_type": "INT"},
    {"name": "boundary_offset", "data_type": "FLOAT_VECTOR"},
]

default_GroupInput = [
    ["atoms_index", "NodeSocketInt"],
    ["species_index", "NodeSocketInt"],
    ["show", "NodeSocketBool"],
    ["select", "NodeSocketInt"],
    ["model_style", "NodeSocketInt"],
    ["scale", "NodeSocketFloat"],
    ["radius_style", "NodeSocketInt"],
    ["boundary_offset", "NodeSocketVector"],
]

default_boundary_datas = {
    "atoms_index": np.ones(0, dtype=int),
    "species_index": np.ones(0, dtype=int),
    "species": np.ones(0, dtype="U4"),
    "positions": np.zeros((0, 3)),
    "scales": np.zeros(0),
    "boundary_offset": np.zeros((0, 3)),
    "model_styles": np.ones(0, dtype=int),
    "shows": np.ones(0, dtype=int),
    "selects": np.ones(0, dtype=int),
}


class Boundary(ObjectGN):
    def __init__(
        self,
        label=None,
        batoms=None,
        boundary=None,
        boundary_datas=None,
        location=(0, 0, 0),
        load=False,
    ):
        """_summary_

        Args:
            label (str, optional):
                _description_. Defaults to None.
            boundary (list, optional):
                search atoms at the boundary.
                Defaults to np.array([[-0.0, 1.0], [-0.0, 1.0], [-0.0, 1.0]]).
            boundary_datas (_type_, optional):
                _description_. Defaults to None.
            batoms (_type_, optional):
                _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        name = "boundary"
        ObjectGN.__init__(self, label, name)
        if not load or not self.loadable():
            if boundary_datas is not None:
                self.build_object(boundary_datas)  # , location=location)
            else:
                self.build_object(default_boundary_datas)  # , location=location)
            if boundary is not None:
                self.boundary = boundary

    def loadable(self):
        """Check loadable or not"""
        # object exist
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            return False
        # batoms exist, and flag is True
        coll = bpy.data.collections.get(self.label)
        if coll is None:
            return False
        return coll.batoms.boundary.flag

    def build_object(self, datas, location=[0, 0, 0], attributes={}):
        """
        build child object and add it to main objects.
        """
        # tstart = time()
        if len(datas["positions"].shape) == 2:
            self._trajectory = {
                "positions": np.array([datas["positions"]]),
            }
            positions = datas["positions"]
        elif len(datas["positions"].shape) == 3:
            self._trajectory = {
                "positions": datas["positions"],
            }
            positions = datas["positions"][0]
        else:
            raise Exception("Shape of positions is wrong!")
        #
        attributes.update(
            {
                "atoms_index": datas["atoms_index"],
                "species_index": datas["species_index"],
                "show": datas["shows"],
                "model_style": datas["model_styles"],
                "select": datas["selects"],
                "scale": datas["scales"],
            }
        )
        name = "%s_boundary" % self.label
        self.delete_obj(name)
        mesh = bpy.data.meshes.new(name)
        mesh.from_pydata(positions, [], [])
        mesh.update()
        obj = bpy.data.objects.new(name, mesh)
        # Add attributes
        for att in default_attributes:
            self.add_attribute(**att)
        self.batoms.coll.objects.link(obj)
        obj.location = location
        obj.batoms.type = "BOUNDARY"
        obj.batoms.label = self.batoms.label
        obj.parent = self.batoms.obj
        self.batoms.coll.batoms.boundary.flag = True
        #
        bpy.context.view_layer.update()
        self.set_attributes(attributes)
        self.init_geometry_node_modifier(default_GroupInput)
        self.build_geometry_node()
        self.set_trajectory()
        # print('boundary: build_object: {0:10.2f} s'.format(time() - tstart))

    def build_geometry_node(self):
        """ """
        from batoms.utils.butils import get_node_by_name

        links = self.gn_node_group.links
        nodes = self.gn_node_group.nodes
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # ------------------------------------------------------------------
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        links.new(GroupInput.outputs["Geometry"], JoinGeometry.inputs["Geometry"])
        links.new(JoinGeometry.outputs["Geometry"], GroupOutput.inputs["Geometry"])
        # ------------------------------------------------------------------
        # transform postions of batoms to boundary
        ObjectBatoms = get_node_by_name(
            nodes, "%s_ObjectBatoms" % self.label, "GeometryNodeObjectInfo"
        )
        ObjectBatoms.inputs["Object"].default_value = self.batoms.obj
        PositionBatoms = get_node_by_name(
            nodes, "%s_PositionBatoms" % (self.label), "GeometryNodeInputPosition"
        )
        TransferBatoms = get_node_by_name(
            nodes, "%s_TransferBatoms" % (self.label), "GeometryNodeSampleIndex"
        )
        TransferBatoms.data_type = "FLOAT_VECTOR"
        links.new(ObjectBatoms.outputs["Geometry"], TransferBatoms.inputs[0])
        links.new(PositionBatoms.outputs["Position"], TransferBatoms.inputs[3])
        links.new(GroupInput.outputs[1], TransferBatoms.inputs["Index"])
        # ------------------------------------------------------------------
        # calculate offset for boundary atoms
        OffsetAttribute = get_node_by_name(
            nodes,
            "%s_boundary_offset" % (self.label),
            "GeometryNodeInputNamedAttribute",
        )
        OffsetAttribute.inputs["Name"].default_value = "boundary_offset"
        OffsetAttribute.data_type = "FLOAT_VECTOR"
        # get arrays of cell
        cell_node = self.get_cell_node(self.gn_node_group)
        cell_node.inputs["Cell"].default_value = self.batoms.cell.obj
        dot_node = self.vector_dot_cell(self.gn_node_group)
        links.new(OffsetAttribute.outputs[0], dot_node.inputs["Vector"])
        for i in range(3):
            links.new(
                cell_node.outputs["A%d" % (i + 1)], dot_node.inputs["Array%d" % (i + 1)]
            )
        # we need one add operation to get the positions with offset
        VectorAdd = get_node_by_name(
            nodes, "%s_VectorAdd" % (self.label), "ShaderNodeVectorMath"
        )
        # ------------------------------------------------------------------
        # add positions with offsets
        VectorAdd.operation = "ADD"
        links.new(TransferBatoms.outputs[2], VectorAdd.inputs[0])
        links.new(dot_node.outputs[0], VectorAdd.inputs[1])
        # set positions
        SetPosition = get_node_by_name(
            nodes, "%s_SetPosition" % self.label, "GeometryNodeSetPosition"
        )
        links.new(GroupInput.outputs["Geometry"], SetPosition.inputs["Geometry"])
        links.new(VectorAdd.outputs[0], SetPosition.inputs["Position"])
        #
        # ------------------------------------------------------------------
        # transform scale of batoms to boundary
        if bpy.app.version_string >= "3.2.0":
            ScaleBatoms = get_node_by_name(
                nodes,
                "%s_ScaleBatoms" % (self.label),
                "GeometryNodeInputNamedAttribute",
            )
            # need to be "FLOAT_VECTOR", because scale is "FLOAT_VECTOR"
            ScaleBatoms.data_type = "FLOAT_VECTOR"
            ScaleBatoms.inputs[0].default_value = "scale"
            TransferScale = get_node_by_name(
                nodes, "%s_TransferScale" % (self.label), "GeometryNodeSampleIndex"
            )
            TransferScale.data_type = "FLOAT_VECTOR"
            links.new(ObjectBatoms.outputs["Geometry"], TransferScale.inputs[0])
            links.new(ScaleBatoms.outputs["Attribute"], TransferScale.inputs[3])
            links.new(GroupInput.outputs[1], TransferScale.inputs["Index"])

    def vector_dot_cell(self, parent_tree):
        """Dot product of a vector (1x3) and a cell (3x3).
        Note, the vector could be a vector field, (Nx3)"""
        from batoms.utils.butils import (
            get_node_by_name,
            create_node_tree,
        )

        default_interface = [
            ["Vector", "NodeSocketVector", "INPUT"],
            ["Array1", "NodeSocketVector", "INPUT"],
            ["Array2", "NodeSocketVector", "INPUT"],
            ["Array3", "NodeSocketVector", "INPUT"],
            ["Vector", "NodeSocketVector", "OUTPUT"],
        ]
        name = "Vector_dot_cell_%s" % (self.label)
        node = get_node_by_name(parent_tree.nodes, name=name, type="GeometryNodeGroup")
        node_tree = create_node_tree(name=name, interface=default_interface)
        node.node_tree = node_tree
        nodes = node_tree.nodes
        links = node_tree.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        #
        SeparateXYZs = []
        CombineXYZs = []
        DotProdcuts = []
        for i in range(3):
            # SeparateXYZ to get the transpose of the cell
            SeparateXYZ = get_node_by_name(
                nodes, f"{self.label}_SeparateXYZ_{i}", "ShaderNodeSeparateXYZ"
            )
            # Dot product of the vector and the array
            DotProdcut = get_node_by_name(
                nodes, f"{self.label}_DotProduct_{i}", "ShaderNodeVectorMath"
            )
            DotProdcut.operation = "DOT_PRODUCT"
            # then combine the vectors, and output to the group output
            CombineXYZ = get_node_by_name(
                nodes, f"{self.label}_CombineXYZ_{i}", "ShaderNodeCombineXYZ"
            )
            links.new(GroupInput.outputs[f"Array{i + 1}"], SeparateXYZ.inputs["Vector"])
            links.new(GroupInput.outputs["Vector"], DotProdcut.inputs[0])
            links.new(CombineXYZ.outputs["Vector"], DotProdcut.inputs[1])
            CombineXYZs.append(CombineXYZ)
            SeparateXYZs.append(SeparateXYZ)
            DotProdcuts.append(DotProdcut)
        for i in range(3):
            for j in range(3):
                links.new(SeparateXYZs[i].outputs[j], CombineXYZs[j].inputs[i])
        # output final vector
        CombineXYZ = get_node_by_name(
            nodes, f"{self.label}_CombineXYZ", "ShaderNodeCombineXYZ"
        )
        for i in range(3):
            links.new(DotProdcuts[i].outputs["Value"], CombineXYZ.inputs[i])
        links.new(CombineXYZ.outputs["Vector"], GroupOutput.inputs["Vector"])
        return node

    def get_cell_node(self, parent_tree):
        """Get the position of the cell."""
        from batoms.utils.butils import (
            get_socket_by_identifier,
            get_node_by_name,
            create_node_tree,
        )

        default_interface = [
            ["Cell", "NodeSocketObject", "INPUT"],
            ["A1", "NodeSocketVector", "OUTPUT"],
            ["A2", "NodeSocketVector", "OUTPUT"],
            ["A3", "NodeSocketVector", "OUTPUT"],
        ]
        name = "Cell_Array_%s" % (self.label)
        node = get_node_by_name(parent_tree.nodes, name=name, type="GeometryNodeGroup")
        node_tree = create_node_tree(name=name, interface=default_interface)
        node.node_tree = node_tree
        nodes = node_tree.nodes
        links = node_tree.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # link the input to parent node
        # ------------------------------------------------------------------
        CellObject = get_node_by_name(
            nodes, f"{self.label}_CellObject", "GeometryNodeObjectInfo"
        )
        links.new(GroupInput.outputs["Cell"], CellObject.inputs["Object"])
        Position = get_node_by_name(
            nodes, "%s_Position" % (self.label), "GeometryNodeInputPosition"
        )
        for i in range(3):
            PositionAtIndex = get_node_by_name(
                nodes, f"{self.label}_PositionAtIndex_{i}", "GeometryNodeSampleIndex"
            )
            PositionAtIndex.data_type = "FLOAT_VECTOR"
            PositionAtIndex.inputs["Index"].default_value = i + 1
            links.new(CellObject.outputs["Geometry"], PositionAtIndex.inputs[0])
            input_socket = get_socket_by_identifier(PositionAtIndex, "Value_Vector")
            links.new(Position.outputs["Position"], input_socket)
            output_socket = get_socket_by_identifier(
                PositionAtIndex, "Value_Vector", type="outputs"
            )
            links.new(output_socket, GroupOutput.inputs["A%d" % (i + 1)])

        return node

    def add_geometry_node(self, spname):
        """ """
        from batoms.utils.butils import get_node_by_name

        links = self.gn_node_group.links
        nodes = self.gn_node_group.nodes
        GroupInput = nodes[0]
        SetPosition = get_node_by_name(nodes, "%s_SetPosition" % self.label)
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        CompareSpecies = get_node_by_name(
            nodes, "CompareFloats_%s_%s" % (self.label, spname), compareNodeType
        )
        CompareSpecies.operation = "EQUAL"
        # CompareSpecies.data_type = 'INT'
        CompareSpecies.inputs[1].default_value = string2Number(spname)
        InstanceOnPoint = get_node_by_name(
            nodes,
            "InstanceOnPoint_%s_%s" % (self.label, spname),
            "GeometryNodeInstanceOnPoints",
        )
        ObjectInfo = get_node_by_name(
            nodes, "ObjectInfo_%s_%s" % (self.label, spname), "GeometryNodeObjectInfo"
        )
        ObjectInfo.inputs["Object"].default_value = self.batoms.species.instancers[
            spname
        ]
        BoolShow = get_node_by_name(
            nodes,
            "BooleanMath_%s_%s_1" % (self.label, spname),
            "FunctionNodeBooleanMath",
        )
        #
        links.new(SetPosition.outputs["Geometry"], InstanceOnPoint.inputs["Points"])
        links.new(GroupInput.outputs[2], CompareSpecies.inputs[0])
        links.new(GroupInput.outputs[3], BoolShow.inputs[0])
        # transfer scale
        TransferScale = get_node_by_name(
            nodes, "%s_TransferScale" % (self.label), "GeometryNodeSampleIndex"
        )
        links.new(TransferScale.outputs[2], InstanceOnPoint.inputs["Scale"])
        links.new(CompareSpecies.outputs[0], BoolShow.inputs[1])
        links.new(BoolShow.outputs["Boolean"], InstanceOnPoint.inputs["Selection"])
        links.new(ObjectInfo.outputs["Geometry"], InstanceOnPoint.inputs["Instance"])
        links.new(InstanceOnPoint.outputs["Instances"], JoinGeometry.inputs["Geometry"])

    def update(self):
        """Main function to update boundary
        1) for each frame, search the boundary list
        2) mearch all boundary list, and find the unique boundary list
        3) calculate all boundary data
        """
        # object_mode()
        # clean_coll_objects(self.coll, 'bond')
        self.hide = False
        if not self.active:
            self.set_arrays(default_boundary_datas)
            return
        positions = self.batoms.positions
        trajectory = self.batoms.get_trajectory()["positions"]
        images = self.batoms.as_ase()
        nframe = self.batoms.nframe
        if nframe == 0:
            images = [images]
            trajectory = np.array([positions])
            nframe = 1
        tstart = time()
        for f in range(nframe):
            # print('update boundary: ', f)
            # print(images[f].positions)
            # use local positions for boundary search
            boundary_list = search_boundary(images[f], self.boundary)
            if f == 0:
                boundary_lists = boundary_list
            else:
                boundary_lists = np.append(boundary_lists, boundary_list, axis=0)
        boundary_lists = np.unique(boundary_lists, axis=0)
        boundary_datas = self.calc_boundary_data(
            boundary_lists, images[0].arrays, trajectory, self.batoms.cell
        )
        # update unit cell

        #
        self.set_arrays(boundary_datas)
        # self.batoms.draw()
        logger.debug("update boundary: {0:10.2f} s".format(time() - tstart))

    def update_geometry_node_instancer(self, spname, instancer):
        """When instances are re-build, we need also update
        the geometry node.

        Args:
            spname (str): name of the species
        """
        from batoms.utils.butils import get_node_by_name

        # update  instancers
        ObjectInfo = get_node_by_name(
            self.gn_node_group.nodes,
            "ObjectInfo_%s_%s" % (self.label, spname),
            "GeometryNodeObjectInfo",
        )
        ObjectInfo.inputs["Object"].default_value = instancer
        logger.debug("update boundary instancer: {}".format(spname))

    def update_gn_cell(self):
        from batoms.utils.butils import get_node_by_name

        # update cell
        cell = self.batoms.cell.array
        # set positions
        nodes = self.gn_node_group.nodes
        for i in range(3):
            tmp = get_node_by_name(
                nodes, "%s_VectorDot%s_%s" % (self.label, i, ""), "ShaderNodeVectorMath"
            )
            tmp.operation = "DOT_PRODUCT"
            tmp.inputs[1].default_value = cell[:, i]

    @property
    def obj(self):
        return self.get_obj()

    def get_obj(self):
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            # , self.batoms.location)
            self.build_object(default_boundary_datas)
        return obj

    @property
    def active(self):
        return self.batoms.coll.batoms.boundary.active

    @active.setter
    def active(self, value):
        self.batoms.coll.batoms.boundary.active = value

    @property
    def boundary(self):
        return self.get_boundary()

    @boundary.setter
    def boundary(self, boundary):
        if boundary is not None:
            if isinstance(boundary, (int, float)):
                boundary = np.array([[-boundary, 1 + boundary]] * 3)
            elif len(boundary) == 3:
                if isinstance(boundary[0], (int, float)):
                    boundary = np.array(
                        [
                            [-boundary[0], 1 + boundary[0]],
                            [-boundary[1], 1 + boundary[1]],
                            [-boundary[2], 1 + boundary[2]],
                        ]
                    )
                elif len(boundary[0]) == 2:
                    boundary = np.array(boundary)
            else:
                raise Exception("Wrong boundary setting!")
            if np.isclose(boundary[:].flatten(), np.array([0, 1, 0, 1, 0, 1])).all():
                self.active = False
            else:
                self.active = True
            self.batoms.coll.batoms.boundary.boundary = boundary[:].flatten()
        else:
            self.active = False
        self.update()
        # self.batoms.draw()

    def get_boundary(self):
        boundary = np.array(self.batoms.coll.batoms.boundary.boundary)
        return boundary.reshape(3, -1)

    def set_arrays(self, arrays):
        """ """
        attributes = self.attributes
        # same length
        dnvert = len(arrays["species_index"]) - len(attributes["species_index"])
        if dnvert > 0:
            self.add_vertices_bmesh(dnvert)
        elif dnvert < 0:
            self.delete_vertices_bmesh(range(-dnvert))
        if len(arrays["positions"]) == 0:
            return
        self.positions = arrays["positions"][0]
        self.set_trajectory(arrays)
        self.update_mesh()
        species_index = [string2Number(sp) for sp in arrays["species"]]
        self.set_attributes(
            {
                "atoms_index": arrays["atoms_index"],
                "species_index": species_index,
                "scale": arrays["scales"],
                "show": arrays["shows"],
                "boundary_offset": arrays["boundary_offset"],
            }
        )
        species = np.unique(arrays["species"])
        for sp in species:
            self.add_geometry_node(sp)

    def get_arrays(self):
        """ """
        # object_mode()
        # tstart = time()
        arrays = self.attributes
        arrays.update({"positions": self.positions})
        # radius
        radius = self.batoms.radius
        arrays.update({"radius": np.zeros(len(self))})
        species = np.array(
            [number2String(i) for i in arrays["species_index"]], dtype="U20"
        )
        arrays["species"] = species
        for sp, value in radius.items():
            mask = np.where(arrays["species"] == sp)
            arrays["radius"][mask] = value
        # size
        arrays["size"] = arrays["radius"] * arrays["scale"]
        # main elements
        main_elements = self.batoms.species.main_elements
        elements = [main_elements[sp] for sp in arrays["species"]]
        arrays.update({"elements": np.array(elements, dtype="U20")})
        # print('get_arrays: %s'%(time() - tstart))
        return arrays

    @property
    def boundary_data(self):
        return self.get_boundary_data()

    def get_boundary_data(self, include_batoms=False):
        """ """
        # check cell voluem
        # if < 1e-6, invalid cell and boundary.
        if self.batoms.cell.volume < 1e-6:
            return None
        arrays = self.arrays
        boundary_data = {
            "positions": arrays["positions"],
            "species": arrays["species"],
            "indices": arrays["atoms_index"],
            "boundary_offset": arrays["boundary_offset"],
        }
        return boundary_data

    @property
    def bondlists(self):
        return self.get_bondlists()

    def get_bondlists(self):
        bondlists = self.batoms.bonds.arrays
        return bondlists

    def get_trajectory(self):
        """ """
        trajectory = {}
        trajectory["positions"] = self.get_obj_trajectory(self.obj)
        return trajectory

    def set_trajectory(self, trajectory=None, frame_start=0):
        if trajectory is None:
            trajectory = self._trajectory
        name = "%s_boundary" % (self.label)
        obj = self.obj
        self.set_shape_key(name, obj, trajectory["positions"], frame_start=frame_start)

    def calc_boundary_data(self, boundary_lists, arrays, positions, cell):
        """ """
        # tstart = time()
        # properties
        model_styles = arrays["model_style"][boundary_lists[:, 0]]
        shows = arrays["show"][boundary_lists[:, 0]]
        selects = arrays["select"][boundary_lists[:, 0]]
        scales = arrays["scale"][boundary_lists[:, 0]]
        species_indexs = arrays["species_index"][boundary_lists[:, 0]]
        species = arrays["species"][boundary_lists[:, 0]]
        # ------------------------------------
        offset_vectors = boundary_lists[:, 1:4]
        offsets = np.dot(offset_vectors, cell)
        # use global positions to build boundary positions
        if len(positions.shape) == 2:
            positions = np.array([positions])
        positions = positions[:, boundary_lists[:, 0]] + offsets
        datas = {
            "atoms_index": np.array(boundary_lists[:, 0]),
            "species_index": species_indexs,
            "species": species,
            "positions": positions,
            # 'offsets':offsets,
            "boundary_offset": offset_vectors,
            "model_styles": model_styles,
            "shows": shows,
            "selects": selects,
            "scales": scales,
        }
        # print('datas: ', datas)
        # print('calc_boundary_data: {0:10.2f} s'.format(time() - tstart))
        return datas

    def __getitem__(self, index):
        return self.boundary[index]

    def __setitem__(self, index, value):
        """Set boundary vectors."""
        boundary = self.boundary
        if isinstance(value, (int, float)):
            boundary = np.array([-boundary, 1 + boundary])
        boundary[index] = value
        self.boundary = boundary

    def __repr__(self) -> str:
        s = self.boundary.__repr__()
        return s

    def as_dict(self):
        """ """
        data = {"array": None}
        if len(self) > 1:
            data["array"] = dict(self.arrays)
        data["boundary"] = self.boundary
        return data


def search_boundary(
    atoms,
    boundary=[[0, 1], [0, 1], [0, 1]],
):
    """Search atoms in the boundary

    Args:
        atoms: _description_
        boundary (list, optional): _description_.
            Defaults to [[0, 1], [0, 1], [0, 1]].

    Returns:
        _type_: _description_
    """
    # tstart = time()
    cell = atoms.cell
    positions = atoms.positions
    species = atoms.arrays["species"]
    if isinstance(boundary, float):
        boundary = [
            [-boundary, 1 + boundary],
            [-boundary, 1 + boundary],
            [-boundary, 1 + boundary],
        ]
    boundary = np.array(boundary)
    # find supercell
    f = np.floor(boundary)
    c = np.ceil(boundary)
    ib = np.array([f[:, 0], c[:, 1]]).astype(int)
    M = np.product(ib[1] - ib[0] + 1)
    # get scaled positions
    positions = np.linalg.solve(complete_cell(cell).T, positions.T).T
    n = len(positions)
    npositions = np.tile(positions, (M - 1,) + (1,) * (len(positions.shape) - 1))
    i0 = 0
    # index
    offsets = np.zeros((M * n, 4), dtype=int)
    ind0 = np.arange(n).reshape(-1, 1)
    species0 = species
    species = []
    # repeat the positions so that
    # it completely covers the boundary
    for m0 in range(ib[0, 0], ib[1, 0] + 1):
        for m1 in range(ib[0, 1], ib[1, 1] + 1):
            for m2 in range(ib[0, 2], ib[1, 2] + 1):
                if m0 == 0 and m1 == 0 and m2 == 0:
                    continue
                i1 = i0 + n
                npositions[i0:i1] += (m0, m1, m2)
                offsets[i0:i1] = np.append(ind0, np.array([[m0, m1, m2]] * n), axis=1)
                species.extend(species0)
                i0 = i1
    # boundary condition
    ind1 = np.where(
        (npositions[:, 0] > boundary[0][0])
        & (npositions[:, 0] < boundary[0][1])  # noqa: W503
        & (npositions[:, 1] > boundary[1][0])  # noqa: W503
        & (npositions[:, 1] < boundary[1][1])  # noqa: W503
        & (npositions[:, 2] > boundary[2][0])  # noqa: W503
        & (npositions[:, 2] < boundary[2][1])  # noqa: W503
    )[0]
    offsets_b = offsets[ind1]
    # print('search boundary: {0:10.2f} s'.format(time() - tstart))
    return offsets_b



import bpy
from bpy.props import (
    IntVectorProperty,
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class CrystalShapeSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="")
    indices: IntVectorProperty(name="Miller indices", size=3, default=[1, 0, 0])
    distance: FloatProperty(
        name="distance", description="Distance from origin", default=1
    )
    color: FloatVectorProperty(
        name="color", size=4, subtype="COLOR", min=0, max=1, default=[0, 0, 1, 0.5]
    )
    crystal: BoolProperty(name="crystal", default=False)
    symmetry: BoolProperty(name="symmetry", default=False)
    slicing: BoolProperty(name="slicing", default=False)
    boundary: BoolProperty(name="boundary", default=False)
    scale: FloatProperty(name="scale", default=1)
    show_edge: BoolProperty(name="show_edge", default=True)
    width: FloatProperty(name="width", default=0.01)

    @property
    def name(self) -> str:
        return "%s-%s-%s" % (self.indices[0], self.indices[1], self.indices[2])

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "name": self.name,
            "material_style": self.material_style,
            "color": self.color[:],
            "indices": self.indices,
            "distance": self.distance,
            "crystal": self.crystal,
            "symmetry": self.symmetry,
            "slicing": self.slicing,
            "boundary": self.boundary,
            "show_edge": self.show_edge,
            "width": self.width,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name        distance  crystal symmetry slicing  show_edge  boundary   edgewidth        \n"
        s += "{0:10s}   {1:1.3f}  {2:8s}  {3:8s}  {4:8s} {5:8s} {6:8s} {7:1.3f}\n".format(
            self.name,
            self.distance,
            str(self.crystal),
            str(self.symmetry),
            str(self.slicing),
            str(self.show_edge),
            str(self.boundary),
            self.width,
        )
        s += "-" * 60 + "\n"
        return s


class CrystalShape(bpy.types.PropertyGroup):
    """This module defines the CrystalShape properties to extend
    Blender’s internal data.

    """

    active: BoolProperty(name="active", default=False)
    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=(
            ("0", "Surface", "Surface"),
            ("1", "Dot", "Dot surface"),
            ("2", "Wireframe", "Use wireframe"),
        ),
        default="0",
    )
    show: BoolProperty(name="show", default=True)
    settings: CollectionProperty(name="CrystalShapeSetting", type=CrystalShapeSetting)

    ui_list_index: IntProperty(name="ui_list_index", default=0)

    def as_dict(self) -> dict:
        setdict = {}
        return setdict



import bpy
from bpy.types import Operator
from bpy.props import (
    IntProperty,
    BoolProperty,
)
import bmesh
from batoms import Batoms
import numpy as np
import logging

logger = logging.getLogger(__name__)


def edit_bond(batoms, indices, order):
    """Change bond order

    Args:
        order (int): The target order
    """
    logger.debug("edit bond")
    species = batoms.arrays["species"]
    bondlists = batoms.bonds.bondlists
    removeH = []
    for i in indices:
        # bond order
        order0 = batoms.bonds[i].order
        # bond atoms idnex
        ai = bondlists[i, 0]
        aj = bondlists[i, 1]
        #
        # how many atoms (nbond) are connected to atoms i and j
        bai0 = np.where((bondlists[:, 0] == ai))[0]
        bai1 = np.where((bondlists[:, 1] == ai))[0]
        baj0 = np.where((bondlists[:, 0] == aj))[0]
        baj1 = np.where((bondlists[:, 1] == aj))[0]
        bai = np.append(bondlists[bai0, 1], bondlists[bai1, 0]).astype(int)
        baj = np.append(bondlists[baj0, 1], bondlists[baj1, 0]).astype(int)
        spsi = species[bai]
        spsj = species[baj]
        indhi = bai[np.where(spsi == "H")[0]]
        indhj = bai[np.where(spsj == "H")[0]]
        # difference between order and target order
        dorder = order - order0
        if dorder >= 0:
            # nbond >= nh,
            # 1. no more new H will be add
            # 2. dh old H will be remove
            dhi = min(len(indhi), dorder)
            dhj = min(len(indhj), dorder)
            removeH.extend(indhi[:dhi].tolist())
            removeH.extend(indhj[:dhj].tolist())
        else:
            # nbond < nh,
            # 1. dh new H will be add
            # 2. no more old H will be remove
            dhi = dorder
            dhj = dorder

    indices.extend(removeH)
    batoms.delete(indices)
    batoms.model_style = 1


class MolecueEditBond(Operator):
    bl_idname = "batoms.molecule_edit_bond"
    bl_label = "Edit bond"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Edit bond"

    order: IntProperty(name="order", default=1, description="order")
    # TODO rotate around bond
    rotate: IntProperty(name="rotate", default=1, description="rotate")
    hydrogen: BoolProperty(
        name="Ajust hydrogen", default=1, description="Ajust hydrogen"
    )

    @classmethod
    def poll(cls, context):
        return context.mode in {"EDIT_MESH"}

    def execute(self, context):
        obj = context.object
        if not obj.batoms.bbond.flag:
            logger.critical("Please select a Batom.")
            return {"FINISHED"}
        data = obj.data
        if data.total_vert_sel > 0:
            bm = bmesh.from_edit_mesh(data)
            indices = [
                s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)
            ]
            self.report({"INFO"}, "%s atoms were replaced" % len(indices))
            batoms = Batoms(label=obj.batoms.label)
            edit_bond(batoms, indices, self.order)
            bpy.context.view_layer.objects.active = batoms.obj
            bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}



import bpy
from batoms.material import create_material
import numpy as np
import logging

logger = logging.getLogger(__name__)
# ========================================================


def draw_cell_curve(coll, verts, label=None):
    """
    Draw unit cell by edge, however, can not be rendered.
    """
    if verts is not None:
        edges = [0, 4, 6, 2, 0, 1, 5, 7, 3, 1]
        crv = bpy.data.curves.new("edge_cell", "CURVE")
        crv.dimensions = "3D"
        spline = crv.splines.new(type="NURBS")
        spline.points.add(len(edges) - 1)
        for p, i in zip(spline.points, edges):
            p.co = np.append(verts[i], [1.0])  # (add nurbs weight)
        cell = bpy.data.objects.new("cell_%s_edge" % label, crv)
        coll.objects.link(cell)


def draw_cylinder(
    name=None,
    datas=[],
    coll=None,
    node_type="Principled BSDF",
    node_inputs=None,
    material_style="plastic",
):
    """
    Draw cylinder.
    """
    from batoms.data.source_data import bond_source

    # materials
    material = create_material(
        name,
        datas["color"],
        node_type=node_type,
        node_inputs=node_inputs,
        material_style=material_style,
    )
    #
    mesh = bpy.data.meshes.new(name)
    obj = bpy.data.objects.new(name, mesh)
    obj.data.materials.append(material)
    if len(datas["centers"]) == 0:
        return obj

    source = bond_source[datas["vertices"]]
    verts, faces = cylinder_mesh_from_vec(
        datas["centers"], datas["normals"], datas["lengths"], datas["width"], source
    )
    mesh = bpy.data.meshes.new(name)
    mesh.from_pydata(verts, [], faces)
    mesh.update()
    mesh.polygons.foreach_set("use_smooth", [True] * len(mesh.polygons))
    obj.data = mesh
    obj.data.materials.append(material)
    #
    for name, inputs in datas["battr_inputs"].items():
        battr = getattr(obj, name)
        for key, value in inputs.items():
            setattr(battr, key, value)
    # bpy.ops.object.shade_smooth()
    coll.objects.link(obj)
    return obj


def draw_surface_from_vertices(name, datas, coll=None, use_smooth=True):
    if len(datas["vertices"]) == 0:
        return
    #
    mesh = bpy.data.meshes.new(name)
    if len(datas["edges"]) > 0:
        mesh.from_pydata(datas["vertices"], datas["edges"], datas["faces"])
    else:
        mesh.from_pydata(datas["vertices"], [], datas["faces"])
    mesh.update()
    mesh.polygons.foreach_set("use_smooth", [use_smooth] * len(mesh.polygons))
    obj = bpy.data.objects.new(name, mesh)
    obj.data = mesh
    #
    for name, inputs in datas["battr_inputs"].items():
        battr = getattr(obj, name)
        for key, value in inputs.items():
            setattr(battr, key, value)
    bpy.ops.object.shade_smooth()
    if coll is not None:
        coll.objects.link(obj)
    return obj


def draw_vertices(name, vertices):
    datas = {
        "vertices": vertices,
        "edges": [],
        "faces": [],
        "color": [0.5, 0.5, 0.5, 1.0],
        "battr_inputs": {},
    }
    coll = bpy.data.collections["Collection"]
    draw_surface_from_vertices(name, datas, coll)


def draw_text(coll_text=None, atoms=None, type=None):
    positions = atoms.positions
    n = len(positions)
    for i in range(n):
        location = positions[i] + np.array([0, 0, 1.0])
        FontCurve = bpy.data.curves.new(type="FONT", name="myFontCurve")
        ob = bpy.data.objects.new("myFontOb", FontCurve)
        if type == 0:
            ob.data.body = "%s" % i
        elif type == 1:
            ob.data.body = "%s" % atoms[i].symbol
        ob.location = location
        coll_text.objects.link(ob)


def draw_2d_slicing(name, datas, coll=None):
    """
    using texture image
    """
    from bpy_extras.image_utils import load_image

    image = load_image(datas["imagename"], check_existing=True)
    bpy.ops.mesh.primitive_plane_add(
        size=1,
        location=datas["location"],
        rotation=datas["rotation"],
    )
    plane = bpy.context.active_object
    bpy.ops.object.mode_set(mode="OBJECT")
    plane.scale = datas["size"]
    plane.data.name = name
    plane.name = name
    material = bpy.data.materials.new(name=name)
    material.use_nodes = True
    material.blend_method = "BLEND"
    node_tree = material.node_tree
    material_output = node_tree.nodes.get("Material Output")
    core_shader = node_tree.nodes.get("Principled BSDF")
    node_texture = node_tree.nodes.new("ShaderNodeTexImage")
    node_texture.image = image
    node_texture.show_texture = True
    node_tree.links.new(core_shader.inputs[0], node_texture.outputs["Color"])
    node_tree.links.new(core_shader.inputs["Alpha"], node_texture.outputs["Alpha"])
    node_tree.links.new(material_output.inputs["Surface"], core_shader.outputs[0])
    plane.data.materials.append(material)
    coll.objects.link(plane)


def clean_node_tree(node_tree):
    """Clear all nodes in a shader node tree except the output.

    Returns the output node
    """
    nodes = node_tree.nodes
    # copy to avoid altering the loop's data source
    for node in list(nodes):
        if not node.type == "OUTPUT_MATERIAL":
            nodes.remove(node)

    return node_tree.nodes[0]


# draw bonds
def bond_source(vertices=12, depth=1.0):
    bpy.ops.mesh.primitive_cylinder_add(vertices=vertices, depth=depth)
    cyli = bpy.context.view_layer.objects.active
    me = cyli.data
    faces = []
    n = len(me.vertices)
    vertices = np.empty(n * 3, dtype=np.float64)
    me.vertices.foreach_get("co", vertices)
    vertices = vertices.reshape((n, 3))
    for poly in me.polygons:
        face = []
        for loop_index in range(poly.loop_start, poly.loop_start + poly.loop_total):
            # print("    Vertex: %d" % me.loops[loop_index].vertex_index)
            face.append(me.loops[loop_index].vertex_index)
        faces.append(face)
    cyli.select_set(True)
    bpy.data.objects.remove(cyli, do_unlink=True)
    n = len(faces[0])
    faces1 = [faces[i] for i in range(len(faces)) if len(faces[i]) == n]
    faces2 = [faces[i] for i in range(len(faces)) if len(faces[i]) != n]
    return vertices, faces1, faces2


# draw bonds


def cylinder2(vertices=16, depth=1.0):
    bpy.ops.mesh.primitive_cylinder_add(vertices=32, depth=1)
    obj = bpy.context.view_layer.objects.active
    me = obj.data
    # select edges for subdivde
    n = len(me.vertices)
    vertices = np.zeros(n * 3, dtype=np.float64)
    me.vertices.foreach_get("co", vertices)
    vertices = vertices.reshape((n, 3))
    #
    me.update()
    m = len(me.edges)
    selects = np.zeros(m, dtype=bool)
    for i in range(m):
        v0 = me.edges[i].vertices[0]
        v1 = me.edges[i].vertices[1]
        center = (vertices[v0] + vertices[v1]) / 2
        # print(i, v0, v1, center)
        if np.isclose(center[2], 0):
            selects[i] = True

    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.select_all(action="DESELECT")
    bpy.ops.object.mode_set(mode="OBJECT")
    me.edges.foreach_set("select", selects)
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.subdivide(number_cuts=1, smoothness=0, fractal_along_normal=0)
    bpy.ops.object.mode_set(mode="OBJECT")
    #
    m = len(me.polygons)
    centers = np.zeros(m * 3, dtype=np.float64)
    me.polygons.foreach_get("center", centers)
    centers = centers.reshape((m, 3))
    return obj


def atom_source():
    bpy.ops.mesh.primitive_uv_sphere_add()  # , segments=32, ring_count=16)
    # bpy.ops.mesh.primitive_cylinder_add()
    sphe = bpy.context.view_layer.objects.active
    me = sphe.data
    verts = []
    faces = []
    for vertices in me.vertices:
        verts.append(np.array(vertices.co))
    for poly in me.polygons:
        face = []
        for loop_index in range(poly.loop_start, poly.loop_start + poly.loop_total):
            # print("    Vertex: %d" % me.loops[loop_index].vertex_index)
            face.append(me.loops[loop_index].vertex_index)
        faces.append(face)
    sphe.select_set(True)
    bpy.ops.object.delete()
    return [verts, faces]


def sphere_mesh_from_instance(centers, radius, source):
    verts = []
    faces = []
    vert0, face0 = source
    nvert = len(vert0)
    nb = len(centers)
    for i in range(nb):
        center = centers[i]
        # r = radius[i]
        # normal = normal/np.linalg.norm(normal)
        for vert in vert0:
            vert = vert * [radius, radius, radius]
            vert += center
            verts.append(vert)
        for face in face0:
            face = [x + i * nvert for x in face]
            faces.append(face)
    return verts, faces


def cylinder_mesh_from_vec(centers, normals, lengths, scale, source):
    from scipy.spatial.transform import Rotation as R

    vert0, face1, face2 = source
    nb = len(centers)
    nvert = len(vert0)
    centers = np.array(centers)
    normals = np.array(normals)
    vert0 = np.array(vert0)
    face1 = np.array(face1)
    face2 = np.array(face2)
    scale = np.array([[scale, scale]] * len(centers))
    scale = np.append(scale, np.array(lengths).reshape(-1, 1), axis=1)
    # print(np.cross([0.0000014159, 0.000001951, 1], normals[0]))
    vec = np.cross([0.0, 0.0, 1], normals) + np.array([1e-6, 0, 0])
    vec = vec / np.linalg.norm(vec, axis=1)[:, None]
    # print(np.arccos(normals[0, 2]*0.999999))
    ang = np.arccos(normals[:, 2] * 0.999999)
    # print(-1*ang[0]*vec[0])
    vec = -1 * (vec.T * ang).T
    # print(R.from_rotvec(vec[0]).as_matrix())
    r = R.from_rotvec(vec)
    matrix = r.as_matrix()
    # print('vert1 0: ', vert0*scale[0])
    verts = np.tile(vert0, (nb, 1, 1))
    verts = verts * scale[:, None]
    # print('matrix: ', verts[0].dot(matrix[0]))
    verts = np.matmul(verts, matrix)
    # print('matrix: ', verts)
    # print(verts.shape)
    # centers = np.tile(centers, (nb, 1))
    # print('center: ', verts[0] + centers[0])
    verts += centers[:, None]
    # print('center: ', verts)
    # faces.append(face)
    verts = verts.reshape(-1, 3)
    # ----------------------------
    nf1 = len(face1[0])
    nf2 = len(face2[0])
    faces1 = np.tile(face1, (nb, 1, 1))
    faces2 = np.tile(face2, (nb, 1, 1))
    offset = np.arange(nb) * nvert
    offset = offset.reshape(-1, 1, 1)
    faces1 = faces1 + offset
    faces1 = faces1.reshape(-1, nf1)
    faces2 = faces2 + offset
    faces2 = faces2.reshape(-1, nf2)
    faces = list(faces1) + list(faces2)
    return verts, faces


def draw_plane(
    name="plane",
    location=(0, 0, -0.01),
    rotation=(0, 0, 0),
    color=(0.2, 0.2, 1.0, 1.0),
    size=200,
    node_type=None,
    node_inputs=None,
    material_style="default",
):
    """
    Draw a plane.
    location: array
    color: array
    size: float
    node_inputs: dict
    material_style: str
    """
    # build materials
    material = create_material(
        name,
        color,
        node_type=node_type,
        node_inputs=node_inputs,
        material_style=material_style,
    )
    # Instantiate a floor plane
    bpy.ops.mesh.primitive_plane_add(size=size, location=location, rotation=rotation)
    obj = bpy.context.object
    obj.data.materials.append(material)
    return obj



import pytest
import sys
from pathlib import Path
import os
from tempfile import TemporaryDirectory

curdir = Path(__file__).parent.resolve()
sys.path.append(curdir.parent.as_posix())


def test_empty_dir():
    from install import _is_empty_dir

    assert _is_empty_dir("/") is False
    with TemporaryDirectory() as tmpdir1:
        assert _is_empty_dir(tmpdir1) is True
        tmpdir2 = Path(tmpdir1) / "test"
        os.makedirs(tmpdir2, exist_ok=True)
        assert _is_empty_dir(tmpdir1) is False
        assert _is_empty_dir(tmpdir2) is True
    with pytest.raises(FileNotFoundError):
        _is_empty_dir(tmpdir1)
        _is_empty_dir(tmpdir2)


def test_default_location_linux(fs):
    """Test requires the `pyfakefs` package's fs fixture"""
    from install import _get_default_locations

    with pytest.raises(NotImplementedError):
        _get_default_locations("linux")


def test_default_location_macos(fs, monkeypatch):
    """Test requires the `pyfakefs` package's fs fixture"""
    monkeypatch.setattr("builtins.input", lambda _: "0")
    from pathlib import Path
    import os
    from install import _get_default_locations

    # macos case 1
    fdn1 = "/Applications/Blender.app/Contents/Resources/3.4"
    fs.create_dir(fdn1)
    assert _get_default_locations("macos") == Path(fdn1)

    # macos case 2: Blender 3.4 and 3.1 both exist, should return 3.4
    fdn2 = "/Applications/Blender.app/Contents/Resources/3.1"
    fs.create_dir(fdn2)
    assert _get_default_locations("macos").name == "3.4"

    fs.rmdir(fdn1)
    with pytest.raises(FileNotFoundError):
        _get_default_locations("macos")
    fs.rmdir(fdn2)

    # macos case 3: both user and application blender
    fdn3 = os.path.expandvars("$HOME/Applications/Blender.app/Contents/Resources/3.4")
    fs.create_dir(fdn1)
    fs.create_dir(fdn3)
    assert _get_default_locations("macos") == Path(fdn1)


def test_blender_bin(fs):
    # Windows
    from pathlib import Path
    from install import _get_blender_bin

    bundle_root = "/c/Program Files/Blender Foundation/Blender 3.4/3.4"
    fake_bin = "/c/Program Files/Blender Foundation/Blender 3.4/blender.exe"
    fs.create_dir(bundle_root)
    fs.create_file(fake_bin)
    assert _get_blender_bin("windows", bundle_root) == Path(fake_bin).resolve()
    # User should not point to the upper level
    with pytest.raises(FileNotFoundError):
        _get_blender_bin("windows", Path(bundle_root).parent)
    fs.rmdir(bundle_root)

    # Linux
    fs.create_dir("./blender/3.4")
    fs.create_file("./blender/blender")
    assert (
        _get_blender_bin("linux", "./blender/3.4")
        == Path("./blender/blender").resolve()  # noqa: W503
    )
    # fs.remove("./blender/blender")
    # fs.rmdir("./blender")

    # Macos
    fs.create_dir("/Applications/Blender.app/Contents/Resources/3.4")
    fs.create_file("/Applications/Blender.app/Contents/MacOS/Blender")
    assert _get_blender_bin("macos", "/Applications/Blender.app/Contents/Resources/3.4")


def test_blender_py_location(monkeypatch):
    def fake_output(*args, **argv):
        output = (
            "Blender 3.2.2 (hash bcfdb14560e7 built 2022-08-02 23:31:44)\n"
            "Read prefs: /Users/ttian/Library/Application Support/Blender/3.2/config/userpref.blend\n"
            "Python binary:  /Applications/Blender.app/Contents/Resources/3.2/python/bin/python3.10\n"
            "\n"
            "Blender quit\n"
        )

        class MockClass:
            stdout = output.encode("utf8")

        return MockClass()

    from pathlib import Path

    monkeypatch.setattr("install._run_blender_multiline_expr", fake_output)
    from install import _get_blender_py

    assert _get_blender_py("test_bin") == Path(
        "/Applications/Blender.app/Contents/Resources/3.2/python/bin/python3.10"
    )


def test_factory_version(monkeypatch):
    def fake_output(*args, **argv):
        output = (
            "Blender 3.4.1 (hash 55485cb379f7 built 2022-12-20 00:46:45)\n"
            "Read prefs: /home/pink/.config/blender/3.4/config/userpref.blend\n"
            "Python Version:  3.10.8 (main, Oct 24 2022, 20:47:11) [GCC 9.3.1 20200408 (Red Hat 9.3.1-2)]\n"
            "Numpy Version:  1.24.2\n"
            "\n"
            "Blender quit\n"
        )

        class MockClass:
            stdout = output.encode("utf8")

        return MockClass()

    monkeypatch.setattr("install._run_blender_multiline_expr", fake_output)
    from install import _get_factory_versions

    py_version, numpy_version = _get_factory_versions("test_bin")
    assert py_version == "3.10.8"
    assert numpy_version == "1.24.2"


def test_blender_version(monkeypatch):
    def fake_output(*args, **argv):
        output = (
            "Blender 3.4.1 (hash 55485cb379f7 built 2022-12-20 00:46:45)\n"
            "Read prefs: /home/pink/.config/blender/3.4/config/userpref.blend\n"
            "\n"
            "Blender quit"
        )

        class MockClass:
            stdout = output.encode("utf8")

        return MockClass()

    monkeypatch.setattr("install._run_process", fake_output)
    from install import _get_blender_version

    assert _get_blender_version("test_bin") == "3.4.1"


def test_rename_dir(fs):
    from install import _rename_dir
    import os

    # case 1: no src
    with pytest.raises(FileNotFoundError):
        _rename_dir("_empty", "_empty")
    # case 2: has src but no dst
    fs.create_dir("source")
    _rename_dir("source", "target")
    assert os.path.isdir("target")
    # case 3: target exists
    fs.create_dir("source1/s1")
    _rename_dir("source1", "target")
    assert os.path.isdir("target")
    assert os.path.isdir("target/s1")
    assert not os.path.isdir("source1")
    # case 4: non-empty target, raise error
    fs.create_dir("source2/s2")
    with pytest.raises(OSError):
        _rename_dir("source2", "target")


def test_symlink(fs):
    from install import _symlink_dir
    import os

    # case 1: no src
    with pytest.raises(FileNotFoundError):
        _symlink_dir("_empty", "_empty")
    # case 2: has src but no dst
    fs.create_dir("source")
    _symlink_dir("source", "target")
    assert os.path.islink("target")
    # case 3: target exists
    fs.create_dir("source1/s1")
    _symlink_dir("source1", "target")
    assert os.path.islink("target")
    assert os.path.isdir("target/s1")
    assert os.path.isdir("source1")
    # case 4: non-empty linked target, will work
    fs.create_dir("source2/s2")
    _symlink_dir("source2", "target")
    assert os.path.isdir("target/s2")
    # case 5: non-empty standard target, will not work
    fs.create_dir("target1/t1")
    with pytest.raises(OSError):
        _symlink_dir("source2", "target1")


def test_binary_file(fs):
    from install import _is_binary_file
    import struct

    fs.create_file("test_plain")
    with open("test_plain", "w") as fd:
        fd.write("sdf")
    assert _is_binary_file("test_plain") is False
    fs.create_file("test_bin")
    with open("test_bin", "wb") as f:
        packed_value = struct.pack("i", 1111)
        f.write(packed_value)
    assert _is_binary_file("test_bin")


def test_conda_name_recognition():
    from install import _is_conda_name_abbrev

    assert _is_conda_name_abbrev("base")
    assert _is_conda_name_abbrev("my-test")
    assert not _is_conda_name_abbrev("./local-env")
    assert not _is_conda_name_abbrev("C:\\local-env")



"""
module for materials
"""
import bpy

material_styles_dict = {
    "default": {
        "type": "Principled BSDF",
        "inputs": {"Metallic": 0.10, "Roughness": 0.20, "IOR": 1.4},
    },
    "metallic": {
        "type": "Principled BSDF",
        "inputs": {"Metallic": 1.0, "Roughness": 0.2, "IOR": 1.4},
    },
    "ceramic": {
        "type": "Principled BSDF",
        "inputs": {"Metallic": 0.02, "Roughness": 0.00, "IOR": 1.4},
    },
    "plastic": {
        "type": "Principled BSDF",
        "inputs": {"Metallic": 0.00, "Roughness": 1.00, "IOR": 1.4},
    },
    "mirror": {
        "type": "Principled BSDF",
        "inputs": {"Metallic": 0.99, "Roughness": 0.01, "IOR": 2.0},
    },
    "glass": {"type": "Glass BSDF", "inputs": {"Roughness": 0.5, "IOR": 1.45}},
    "emission": {"type": "Emission", "inputs": {"Strength": 5.0}},
}


def create_material(
    name,
    color,
    node_type="Principled BSDF",
    node_inputs=None,
    material_style="default",
    backface_culling=True,
    vertex_color=None,
    color_by_attribute=None,
    alpha=None,
):
    """
    Creat material
    """
    if node_inputs is None:
        node_inputs = material_styles_dict[material_style]["inputs"]
        node_type = material_styles_dict[material_style]["type"]
    material = bpy.data.materials.new(name)
    material.diffuse_color = color
    for key, value in node_inputs.items():
        if hasattr(material, key):
            setattr(material, key, value)
    material.blend_method = "BLEND"
    material.use_nodes = True
    #
    nodes = material.node_tree.nodes
    mat_links = material.node_tree.links
    if node_type == "Glass BSDF":
        nodes.new("ShaderNodeBsdfGlass")
    elif node_type == "Emission":
        nodes.new("ShaderNodeEmission")
    node = nodes[node_type]
    node.inputs[0].default_value = color
    # Alpha transpancy
    if "Alpha" in node.inputs:
        if alpha is not None:
            node.inputs["Alpha"].default_value = alpha
        else:
            node.inputs["Alpha"].default_value = color[3]

    for key, value in node_inputs.items():
        node.inputs[key].default_value = value
    if backface_culling:
        material.use_backface_culling = True
    material.show_transparent_back = False

    # vertex color
    if vertex_color not in [None, "None"]:
        if bpy.data.version > (3, 0, 0):
            vcol = nodes.new(type="ShaderNodeVertexColor")
        else:
            vcol = nodes.new(type="ShaderNodeVertexColor")
        vcol.layer_name = vertex_color
        mat_links.new(vcol.outputs["Color"], node.inputs["Base Color"])
        mat_links.new(vcol.outputs["Alpha"], node.inputs["Alpha"])
    # color by attribute
    if color_by_attribute is not None:
        attr_node = nodes.new(type="ShaderNodeAttribute")
        attr_node.attribute_name = color_by_attribute["attribute_name"]
        if "ValToRGB" in color_by_attribute:
            color_ramp = nodes.new(type="ShaderNodeValToRGB")
            color_ramp.color_ramp.elements[0].color = color_by_attribute["ValToRGB"][0]
            color_ramp.color_ramp.elements[1].color = color_by_attribute["ValToRGB"][1]
            mat_links.new(attr_node.outputs["Fac"], color_ramp.inputs["Fac"])
            mat_links.new(color_ramp.outputs["Color"], node.inputs["Base Color"])
            mat_links.new(color_ramp.outputs["Alpha"], node.inputs["Alpha"])
        else:
            mat_links.new(attr_node.outputs["Color"], node.inputs["Base Color"])
        # mat_links.new(vcol.outputs['Alpha'], node.inputs['Alpha'])

    return material


def clean_node_tree(nodes):
    """
    Clear all nodes except the output.
    """
    for node in list(nodes):
        if not node.type == "OUTPUT_MATERIAL":
            nodes.remove(node)



"""
"""
from batoms import Batoms
import logging

logger = logging.getLogger(__name__)


def pubchem_search(cid):
    from ase.data.pubchem import pubchem_search

    # import ssl
    # ssl._create_default_https_context = ssl._create_unverified_context
    data = pubchem_search(cid=cid)
    batoms = Batoms(label="cid_%s" % cid, from_ase=data.atoms)
    return batoms



import bpy
from bpy_extras.io_utils import ImportHelper
from bpy.props import BoolProperty, StringProperty, EnumProperty
from batoms import Batoms
from batoms.ops.base import OperatorBatoms
from batoms.internal_data.bpy_data import get_volumetric_data


class VolumetricDataAdd(OperatorBatoms, ImportHelper):
    bl_idname = "batoms.volumetric_data_add"
    bl_label = "Add volumetric data"
    bl_description = "Add volumetric data to a Batoms"

    filename_ext = ".cube"

    name: StringProperty(name="name", default="", description="name to be added")

    def execute(self, context):
        from ase.io.cube import read_cube_data
        import numpy as np

        if self.name == "":
            return {"FINISHED"}
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        if ".cube" in self.filepath:
            data, atoms = read_cube_data(bpy.path.abspath(self.filepath))
            cell = atoms.cell
        elif "CHGCAR" in self.filepath:
            from pymatgen.io.vasp.outputs import VolumetricData

            poscar, data, data_aug = VolumetricData.parse_file(self.filepath)
            data = data["total"]
            cell = poscar.structure.lattice.matrix
        # check demsion
        assert np.isclose(cell, batoms.cell).all()
        batoms.volumetric_data[self.name] = data
        context.view_layer.objects.active = obj
        self.report({"INFO"}, "Volumetric data {} is added.".format(self.name))
        return {"FINISHED"}


class VolumetricDataRemove(OperatorBatoms):
    bl_idname = "batoms.volumetric_data_remove"
    bl_label = "Remove volumetric data"
    bl_description = "Remove volumetric data to a Batoms"

    name: StringProperty(name="name", default="C", description="name to be removed")

    all: BoolProperty(name="all", default=False, description="Remove all data")

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.volumetric_data.remove((self.name))
        context.view_layer.objects.active = obj
        self.report({"INFO"}, "Volumetric data {} is removed.".format(self.name))
        return {"FINISHED"}


class VolumetricDataCreate(OperatorBatoms):
    bl_idname = "batoms.volumetric_data_create"
    bl_label = "Create volumetric data"
    bl_description = "Create volumetric data to a Batoms. Add or minus."

    name: StringProperty(
        name="name", default="", description="Volumetric data to be Created."
    )

    select_data1: EnumProperty(
        name="select_data1",
        items=get_volumetric_data,
        description="Volumetric data to be used.",
        default=0,
    )

    select_data2: EnumProperty(
        name="select_data2",
        items=get_volumetric_data,
        description="Volumetric data to be used.",
        default=0,
    )

    operator: EnumProperty(
        name="operator",
        items=[("Add", "Add", "", 0), ("Minus", "Minus", "", 1)],
        description="Volumetric data to be used.",
        default=0,
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        if len(batoms.volumetric_data) < 2:
            self.report(
                {"WARNING"},
                "Not enough volumetric data to be used for {}.".format(self.name),
            )
            return {"FINISHED"}
        if self.name in batoms.volumetric_data.keys():
            self.name = "{}_1".format(self.name)
        if self.operator.upper() == "ADD":
            batoms.volumetric_data[self.name] = (
                batoms.volumetric_data[self.select_data1]
                + batoms.volumetric_data[self.select_data2]  # noqa: W503
            )
        else:
            batoms.volumetric_data[self.name] = (
                batoms.volumetric_data[self.select_data1]
                - batoms.volumetric_data[self.select_data2]  # noqa: W503
            )
        context.view_layer.objects.active = obj
        self.report({"INFO"}, "Volumetric data {} is created.".format(self.name))
        return {"FINISHED"}



import bpy
from bpy.types import (
    Panel,
)
from bpy.props import (
    FloatProperty,
    BoolProperty,
)
from batoms.utils.butils import get_selected_vertices
from batoms.utils.attribute import get_mesh_attribute_bmesh, set_mesh_attribute_bmesh


class Batom_PT_prepare(Panel):
    bl_label = "Batom"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "BATOMS_PT_Batom"

    # @classmethod
    # def poll(cls, context):
    #     obj = context.object
    #     if obj:
    #         return obj.batoms.type == 'BATOMS' and obj.mode != 'EDIT'
    #     else:
    #         return False

    def draw(self, context):
        layout = self.layout
        batom = context.scene.batoms.batom
        layout.prop(batom, "scale")
        layout.prop(batom, "show")
        layout.prop(batom, "bond")


# ---------------------------------------------------
def get_selected_batom():
    """Get the indices and the object of the selected vertices"""
    # bpy.ops.object.mode_set(mode="OBJECT")
    context = bpy.context
    if context.object and context.object.batoms.type != "OTHER":
        indices = get_selected_vertices(context.object)
        if len(indices) > 0:
            return context.object, indices
        else:
            return context.object, []
    bpy.ops.object.mode_set(mode="EDIT")
    return None, []


def set_object_attr(key, value):
    """set the attribute of the object"""
    # bpy.ops.object.mode_set(mode="OBJECT")
    obj, indices = get_selected_batom()
    if obj is not None and len(indices) > 0:
        for index in indices:
            set_mesh_attribute_bmesh(obj, key, value, index=index)
        # bpy.context.view_layer.objects.active = batom.obj
    bpy.ops.object.mode_set(mode="EDIT")


def get_scale(self):
    obj, indices = get_selected_batom()
    if obj is not None and len(indices) > 0:
        return get_mesh_attribute_bmesh(obj, "scale", index=indices[0])[0]
    else:
        return 0


def set_scale(self, value):
    self["scale"] = value
    set_object_attr("scale", value)


def get_show(self):
    obj, indices = get_selected_batom()
    if obj is not None and len(indices) > 0:
        return bool(get_mesh_attribute_bmesh(obj, "show", index=indices[0])[0])
    else:
        return 0


def set_show(self, value):
    self["show"] = value
    set_object_attr("show", value)


class BatomProperties(bpy.types.PropertyGroup):
    def Callback_modify_size(self, context):
        batom = bpy.context.scene.batoms.batom
        size = batom.size
        bpy.ops.batoms.batom_modify(key="size", size=size)

    def Callback_modify_bond(self, context):
        batom = bpy.context.scene.batoms.batom
        bond = batom.bond
        bpy.ops.batoms.batom_modify(key="bond", bond=bond)

    scale: FloatProperty(
        name="scale",
        default=0.6,
        min=0,
        soft_max=2,
        description="scale",
        get=get_scale,
        set=set_scale,
    )

    size: FloatProperty(
        name="size",
        default=1.5,
        min=0,
        soft_max=4,
        description="size",
        update=Callback_modify_size,
    )

    bond: BoolProperty(
        name="Bond", default=True, description="bond", update=Callback_modify_bond
    )

    show: BoolProperty(
        name="Show", default=True, description="show", get=get_show, set=set_show
    )
